from __future__ import unicode_literals

import importlib
import json
import random
import re
import string
from datetime import datetime
from typing import List
from unittest.mock import Mock, patch

import django_comments
import pytest
from django.contrib.auth.models import AnonymousUser, User
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.http import Http404, JsonResponse
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django_comments.views.comments import CommentPostBadRequest
from django_comments_ink import signals, signed, views
from django_comments_ink.conf import settings
from django_comments_ink.models import CommentReaction, InkComment
from django_comments_ink.tests.models import Article, Diary
from rest_framework import status
from rest_framework.exceptions import ErrorDetail, PermissionDenied

request_factory = RequestFactory()


def post_article_comment(data, article, auth_user=None):
    request = request_factory.post(
        reverse(
            "article-detail",
            kwargs={
                "year": article.publish.year,
                "month": article.publish.month,
                "day": article.publish.day,
                "slug": article.slug,
            },
        ),
        data=data,
        follow=True,
    )
    if auth_user:
        request.user = auth_user
    else:
        request.user = AnonymousUser()
    request._dont_enforce_csrf_checks = True
    return views.post(request)


def post_diary_comment(data, diary_entry, auth_user=None):
    request = request_factory.post(
        reverse(
            "diary-detail",
            kwargs={
                "year": diary_entry.publish.year,
                "month": diary_entry.publish.month,
                "day": diary_entry.publish.day,
            },
        ),
        data=data,
        follow=True,
    )
    if auth_user:
        request.user = auth_user
    else:
        request.user = AnonymousUser()
    request._dont_enforce_csrf_checks = True
    return views.post(request)
    # return comments.post_comment(request)


def confirm_comment_url(key, follow=True):
    request = request_factory.get(
        reverse("comments-ink-confirm", kwargs={"key": key}), follow=follow
    )
    request.user = AnonymousUser()
    return views.confirm(request, key)


app_model_options_mock = {"tests.article": {"who_can_post": "users"}}


class OnCommentWasPostedTestCase(TestCase):
    def setUp(self):
        self.patcher = patch("django_comments_ink.views.utils.send_mail")
        self.mock_mailer = self.patcher.start()
        self.article = Article.objects.create(
            title="October", slug="october", body="What I did on October..."
        )
        self.form = django_comments.get_form()(self.article)
        self.user = AnonymousUser()

    def tearDown(self):
        self.patcher.stop()

    def post_valid_data(self, auth_user=None, response_code=302):
        data = {
            "name": "Bob",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
            "next": reverse("comments-ink-sent"),
        }
        data.update(self.form.initial)
        response = post_article_comment(data, self.article, auth_user)
        self.assertEqual(response.status_code, response_code)
        if response.status_code == 302:
            self.assertTrue(response.url.startswith("/comments/sent/?c="))

    def post_invalid_data(
        self, auth_user=None, response_code=302, remove_fields: List[str] = []
    ):
        data = {
            "name": "Bob",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
            "next": reverse("comments-ink-sent"),
        }
        data.update(self.form.initial)
        if len(remove_fields):
            for field_name in remove_fields:
                data.pop(field_name)
        response = post_article_comment(data, self.article, auth_user)
        self.assertEqual(response.status_code, response_code)
        if response.status_code == 302:
            self.assertTrue(response.url.startswith("/comments/sent/?c="))

    def test_post_as_authenticated_user(self):
        self.user = User.objects.create_user("bob", "bob@example.com", "pwd")
        self.assertTrue(self.mock_mailer.call_count == 0)
        self.post_valid_data(auth_user=self.user)
        # no confirmation email sent as user is authenticated
        self.assertTrue(self.mock_mailer.call_count == 0)

    def test_post_as_authenticated_user_without_name_nor_email(self):
        data = {
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
        }
        self.user = User.objects.create_user("bob", "bob@example.com", "pwd")
        self.assertTrue(self.mock_mailer.call_count == 0)
        self.post_invalid_data(
            auth_user=self.user, remove_fields=["name", "email"]
        )
        # no confirmation email sent as user is authenticated via self.user.
        self.assertTrue(self.mock_mailer.call_count == 0)

    # def test_post_comment_form_without_content_type(self):
    #     self.user = User.objects.create_user("bob", "bob@example.com", "pwd")
    #     self.assertTrue(self.mock_mailer.call_count == 0)
    #     self.post_invalid_data(
    #         auth_user=self.user,
    #         response_code=400,
    #         remove_fields=['content_type']
    #     )

    def test_confirmation_email_is_sent(self):
        self.assertTrue(self.mock_mailer.call_count == 0)
        self.post_valid_data()
        self.assertTrue(self.mock_mailer.call_count == 1)

    @patch.multiple(
        "django_comments_ink.conf.settings",
        COMMENTS_INK_APP_MODEL_OPTIONS=app_model_options_mock,
    )
    def test_post_as_visitor_when_only_users_can_post(self):
        self.assertTrue(self.mock_mailer.call_count == 0)
        self.post_valid_data(response_code=400)
        self.assertTrue(self.mock_mailer.call_count == 0)


class ConfirmCommentTestCase(TestCase):
    def setUp(self):
        patcher = patch("django_comments_ink.views.utils.send_mail")
        self.mock_mailer = patcher.start()
        # Create random string so that it's harder for zlib to compress
        content = "".join(random.choice(string.printable) for _ in range(6096))
        self.article = Article.objects.create(
            title="September",
            slug="september",
            body="In September..." + content,
        )
        self.form = django_comments.get_form()(self.article)
        data = {
            "name": "Bob",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal iene kleine...",
            "next": reverse("comments-ink-sent"),
        }
        data.update(self.form.initial)
        post_article_comment(data, self.article)
        self.assertTrue(self.mock_mailer.call_count == 1)
        self.key = str(
            re.search(
                r"http://.+/confirm/(?P<key>[\S]+)/",
                self.mock_mailer.call_args[0][1],
            ).group("key")
        )
        self.addCleanup(patcher.stop)

    def test_confirm_url_is_short_enough(self):
        # Tests that the length of the confirm url's length isn't
        # dependent on the article length.
        l = len(reverse("comments-ink-confirm", kwargs={"key": self.key}))
        # print("\nXXXXXXXXXXX:", l)
        self.assertLessEqual(l, 4096, "Urls can only be a max of 4096")

    def test_400_on_bad_signature(self):
        response = confirm_comment_url(self.key[:-1])
        self.assertEqual(response.status_code, 400)

    def test_consecutive_confirmation_url_visits_doesnt_fail(self):
        # test that consecutives visits to the same confirmation URL produce
        # an Http 404 code, as the comment has already been verified in the
        # first visit
        response = confirm_comment_url(self.key)
        self.assertEqual(response.status_code, 302)
        confirm_comment_url(self.key)
        self.assertEqual(response.status_code, 302)

    def test_signal_receiver_may_discard_the_comment(self):
        # test that receivers of signal confirmation_received may return False
        # and thus rendering a template_discarded output
        def on_signal(sender, comment, request, **kwargs):
            return False

        self.assertEqual(self.mock_mailer.call_count, 1)  # sent during setUp
        signals.confirmation_received.connect(on_signal)
        response = confirm_comment_url(self.key)
        # mailing avoided by on_signal:
        self.assertEqual(self.mock_mailer.call_count, 1)
        self.assertTrue(response.content.find(b"Comment discarded") > -1)

    def test_comment_is_created_and_view_redirect(self):
        # testing that visiting a correct confirmation URL creates a InkComment
        # and redirects to the article detail page
        Site.objects.get_current().domain = "testserver"  # django bug #7743
        response = confirm_comment_url(self.key, follow=False)
        data = signed.loads(self.key, extra_key=settings.COMMENTS_INK_SALT)
        try:
            comment = InkComment.objects.get(
                content_type=data["content_type"],
                user_name=data["user_name"],
                user_email=data["user_email"],
                submit_date=data["submit_date"],
            )
        except:
            comment = None
        self.assertTrue(comment is not None)
        self.assertEqual(response.url, comment.get_absolute_url())

    def test_notify_comment_followers(self):
        # send a couple of comments to the article with followup=True and check
        # that when the second comment is confirmed a followup notification
        # email is sent to the user who sent the first comment
        self.assertEqual(self.mock_mailer.call_count, 1)
        confirm_comment_url(self.key)
        # no comment followers yet:
        self.assertEqual(self.mock_mailer.call_count, 1)
        # send 2nd comment
        self.form = django_comments.get_form()(self.article)
        data = {
            "name": "Alice",
            "email": "alice@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
            "next": reverse("comments-ink-sent"),
        }
        data.update(self.form.initial)
        response = post_article_comment(data, article=self.article)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith("/comments/sent/?c="))
        self.assertEqual(self.mock_mailer.call_count, 2)
        self.key = re.search(
            r"http://.+/confirm/(?P<key>[\S]+)/",
            self.mock_mailer.call_args[0][1],
        ).group("key")
        confirm_comment_url(self.key)
        self.assertEqual(self.mock_mailer.call_count, 3)
        self.assertTrue(self.mock_mailer.call_args[0][3] == ["bob@example.com"])
        self.assertTrue(
            self.mock_mailer.call_args[0][1].find(
                "There is a new comment following up yours."
            )
            > -1
        )

    @patch.multiple(
        "django_comments_ink.conf.settings", COMMENTS_INK_SEND_HTML_EMAIL=False
    )
    def test_notify_comment_followers_with_SEND_HTML_EMAIL_eq_False(self):
        # send a couple of comments to the article with followup=True and check
        # that when the second comment is confirmed a followup notification
        # email is sent to the user who sent the first comment
        self.assertEqual(self.mock_mailer.call_count, 1)
        confirm_comment_url(self.key)
        # no comment followers yet:
        self.assertEqual(self.mock_mailer.call_count, 1)
        # send 2nd comment
        self.form = django_comments.get_form()(self.article)
        data = {
            "name": "Alice",
            "email": "alice@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
            "next": reverse("comments-ink-sent"),
        }
        data.update(self.form.initial)
        response = post_article_comment(data, article=self.article)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith("/comments/sent/?c="))
        self.assertEqual(self.mock_mailer.call_count, 2)
        self.key = re.search(
            r"http://.+/confirm/(?P<key>[\S]+)/",
            self.mock_mailer.call_args[0][1],
        ).group("key")
        confirm_comment_url(self.key)
        self.assertEqual(self.mock_mailer.call_count, 3)
        self.assertTrue(self.mock_mailer.call_args[0][3] == ["bob@example.com"])
        self.assertTrue(
            self.mock_mailer.call_args[0][1].find(
                "There is a new comment following up yours."
            )
            > -1
        )
        self.assertEqual(self.mock_mailer.call_args[1], {"html": None})

    def test_notify_followers_dupes(self):
        # first of all confirm Bob's comment otherwise it doesn't reach DB
        confirm_comment_url(self.key)
        # then put in play pull-request-15's assert...
        # https://github.com/danirus/django-comments-xtd/pull/15
        diary = Diary.objects.create(body="Lorem ipsum", allow_comments=True)
        self.assertEqual(diary.pk, self.article.pk)

        self.form = django_comments.get_form()(diary)
        data = {
            "name": "Charlie",
            "email": "charlie@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
            "next": reverse("comments-ink-sent"),
        }
        data.update(self.form.initial)
        response = post_diary_comment(data, diary_entry=diary)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith("/comments/sent/?c="))
        self.key = str(
            re.search(
                r"http://.+/confirm/(?P<key>[\S]+)/",
                self.mock_mailer.call_args[0][1],
            ).group("key")
        )
        # 1) confirmation for Bob (sent in `setUp()`)
        # 2) confirmation for Charlie
        self.assertEqual(self.mock_mailer.call_count, 2)
        response = confirm_comment_url(self.key)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith("/comments/cr/"))
        self.assertEqual(self.mock_mailer.call_count, 2)

        self.form = django_comments.get_form()(self.article)
        data = {
            "name": "Alice",
            "email": "alice@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal iene kleine...",
            "next": reverse("comments-ink-sent"),
        }
        data.update(self.form.initial)
        response = post_article_comment(data, article=self.article)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith("/comments/sent/?c="))
        self.assertEqual(self.mock_mailer.call_count, 3)
        self.key = re.search(
            r"http://.+/confirm/(?P<key>[\S]+)/",
            self.mock_mailer.call_args[0][1],
        ).group("key")
        confirm_comment_url(self.key)
        self.assertEqual(self.mock_mailer.call_count, 4)
        self.assertTrue(self.mock_mailer.call_args[0][3] == ["bob@example.com"])
        self.assertTrue(
            self.mock_mailer.call_args[0][1].find(
                "There is a new comment following up yours."
            )
            > -1
        )

    def test_no_notification_for_same_user_email(self):
        # test that a follow-up user_email don't get a notification when
        # sending another email to the thread
        self.assertEqual(self.mock_mailer.call_count, 1)
        confirm_comment_url(self.key)  # confirm Bob's comment
        # no comment followers yet:
        self.assertEqual(self.mock_mailer.call_count, 1)
        # send Bob's 2nd comment
        self.form = django_comments.get_form()(self.article)
        data = {
            "name": "Alice",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Bob's comment he shouldn't get notified about",
            "next": reverse("comments-ink-sent"),
        }
        data.update(self.form.initial)
        response = post_article_comment(data, self.article)
        self.assertEqual(self.mock_mailer.call_count, 2)
        self.key = re.search(
            r"http://.+/confirm/(?P<key>[\S]+)/",
            self.mock_mailer.call_args[0][1],
        ).group("key")
        confirm_comment_url(self.key)
        self.assertEqual(self.mock_mailer.call_count, 2)


class ReplyNoCommentTestCase(TestCase):
    def test_reply_non_existing_comment_raises_404(self):
        response = self.client.get(
            reverse("comments-ink-reply", kwargs={"cid": 1})
        )
        self.assertContains(response, "404", status_code=404)


class ReplyCommentTestCase(TestCase):
    def setUp(self):
        article = Article.objects.create(
            title="September",
            slug="september",
            body="What I did on September...",
        )
        article_ct = ContentType.objects.get(app_label="tests", model="article")
        site = Site.objects.get(pk=1)

        # post Comment 1 to article, level 0
        InkComment.objects.create(
            content_type=article_ct,
            object_pk=article.id,
            content_object=article,
            site=site,
            comment="comment 1 to article",
            submit_date=datetime.now(),
        )

        # post Comment 2 to article, level 1
        InkComment.objects.create(
            content_type=article_ct,
            object_pk=article.id,
            content_object=article,
            site=site,
            comment="comment 1 to comment 1",
            submit_date=datetime.now(),
            parent_id=1,
        )

        # post Comment 3 to article, level 2 (max according to test settings)
        InkComment.objects.create(
            content_type=article_ct,
            object_pk=article.id,
            content_object=article,
            site=site,
            comment="comment 1 to comment 1",
            submit_date=datetime.now(),
            parent_id=2,
        )

    def test_reply_view(self):
        response = self.client.get(
            reverse("comments-ink-reply", kwargs={"cid": 3})
        )
        self.assertEqual(response.status_code, 200)

    @patch.multiple(
        "django_comments_ink.conf.settings", COMMENTS_INK_MAX_THREAD_LEVEL=2
    )
    def test_not_allow_threaded_reply_raises_403(self):
        response = self.client.get(
            reverse("comments-ink-reply", kwargs={"cid": 3})
        )
        self.assertEqual(response.status_code, 403)

    @patch.multiple(
        "django_comments_ink.conf.settings",
        COMMENTS_INK_APP_MODEL_OPTIONS=app_model_options_mock,
    )
    def test_reply_as_visitor_when_only_users_can_post(self):
        response = self.client.get(
            reverse("comments-ink-reply", kwargs={"cid": 1})
        )
        self.assertEqual(response.status_code, 302)  # Redirect to login.
        self.assertTrue(response.url.startswith(settings.LOGIN_URL))


class MuteFollowUpTestCase(TestCase):
    def setUp(self):
        # Creates an article and send two comments to the article with
        # follow-up notifications. First comment doesn't have to send any
        #  notification.
        # Second comment has to send one notification (to Bob).
        patcher = patch("django_comments_ink.views.utils.send_mail")
        self.mock_mailer = patcher.start()
        self.article = Article.objects.create(
            title="September", slug="september", body="John's September"
        )
        self.form = django_comments.get_form()(self.article)

        # Bob sends 1st comment to the article with follow-up
        data = {
            "name": "Bob",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Nice September you had...",
            "next": reverse("comments-ink-sent"),
        }
        data.update(self.form.initial)
        response = post_article_comment(data, self.article)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith("/comments/sent/?c="))
        self.assertTrue(self.mock_mailer.call_count == 1)
        bobkey = str(
            re.search(
                r"http://.+/confirm/(?P<key>[\S]+)/",
                self.mock_mailer.call_args[0][1],
            ).group("key")
        )
        confirm_comment_url(bobkey)  # confirm Bob's comment

        # Alice sends 2nd comment to the article with follow-up
        data = {
            "name": "Alice",
            "email": "alice@example.com",
            "followup": True,
            "reply_to": 1,
            "level": 1,
            "order": 1,
            "comment": "Yeah, great photos",
            "next": reverse("comments-ink-sent"),
        }
        data.update(self.form.initial)
        response = post_article_comment(data, self.article)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith("/comments/sent/?c="))
        self.assertTrue(self.mock_mailer.call_count == 2)
        alicekey = str(
            re.search(
                r"http://.+/confirm/(?P<key>[\S]+)/",
                self.mock_mailer.call_args[0][1],
            ).group("key")
        )
        confirm_comment_url(alicekey)  # confirm Alice's comment

        # Bob receives a follow-up notification
        self.assertTrue(self.mock_mailer.call_count == 3)
        self.bobs_mutekey = str(
            re.search(
                r"http://.+/mute/(?P<key>[\S]+)/",
                self.mock_mailer.call_args[0][1],
            ).group("key")
        )
        self.addCleanup(patcher.stop)

    def get_mute_followup_url(self, key):
        request = request_factory.get(
            reverse("comments-ink-mute", kwargs={"key": key}), follow=True
        )
        request.user = AnonymousUser()
        response = views.mute(request, key)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.content.find(b"Comment thread muted") > -1)
        return response

    def test_mute_followup_notifications(self):
        # Bob's receive a notification and click on the mute link to
        # avoid additional comment messages on the same article.
        self.get_mute_followup_url(self.bobs_mutekey)
        # Alice sends 3rd comment to the article with follow-up
        data = {
            "name": "Alice",
            "email": "alice@example.com",
            "followup": True,
            "reply_to": 2,
            "level": 1,
            "order": 1,
            "comment": "And look at this and that...",
            "next": reverse("comments-ink-sent"),
        }
        data.update(self.form.initial)
        response = post_article_comment(data, self.article)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith("/comments/sent/?c="))
        # Alice confirms her comment...
        self.assertTrue(self.mock_mailer.call_count == 4)
        alicekey = str(
            re.search(
                r"http://.+/confirm/(?P<key>[\S]+)/",
                self.mock_mailer.call_args[0][1],
            ).group("key")
        )
        confirm_comment_url(alicekey)  # confirm Alice's comment
        # Alice confirmed her comment, but this time Bob won't receive any
        # notification, neither do Alice being the sender
        self.assertTrue(self.mock_mailer.call_count == 4)

    def test_mute_followup_notifications_with_wrong_key(self):
        # Bob's receive a notification and click on the mute link to
        # avoid additional comment messages on the same article.
        # But we use a wrong key (just remove a few bits from the tail).
        key = self.bobs_mutekey[:-2]
        request = request_factory.get(
            reverse("comments-ink-mute", kwargs={"key": key}), follow=True
        )
        request.user = AnonymousUser()
        response = views.mute(request, key)
        self.assertEqual(response.status_code, 400)
        self.assertTrue(response.content.find(b"Bad Request") > -1)

    def test_muting_twice_raise_http_404(self):
        # Mute one time.
        self.get_mute_followup_url(self.bobs_mutekey)
        # Mute a second time.
        request = request_factory.get(
            reverse("comments-ink-mute", kwargs={"key": self.bobs_mutekey}),
            follow=True,
        )
        request.user = AnonymousUser()
        with self.assertRaises(Http404):
            views.mute(request, self.bobs_mutekey)


class HTMLDisabledMailTestCase(TestCase):
    def setUp(self):
        # Create an article and send a comment. Test method will chech headers
        # to see wheter messages has multiparts or not.
        patcher = patch("django_comments_ink.views.utils.send_mail")
        self.mock_mailer = patcher.start()
        self.article = Article.objects.create(
            title="September", slug="september", body="John's September"
        )
        self.form = django_comments.get_form()(self.article)

        # Bob sends 1st comment to the article with follow-up
        self.data = {
            "name": "Bob",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Nice September you had...",
            "next": reverse("comments-ink-sent"),
        }
        self.data.update(self.form.initial)

    @patch.multiple(
        "django_comments_ink.conf.settings", COMMENTS_INK_SEND_HTML_EMAIL=False
    )
    def test_mail_does_not_contain_html_part(self):
        with patch.multiple(
            "django_comments_ink.conf.settings",
            COMMENTS_INK_SEND_HTML_EMAIL=False,
        ):
            response = post_article_comment(self.data, self.article)
            self.assertEqual(response.status_code, 302)
            self.assertTrue(response.url.startswith("/comments/sent/?c="))
            self.assertTrue(self.mock_mailer.call_count == 1)
            self.assertTrue(self.mock_mailer.call_args[1]["html"] is None)

    def test_mail_does_contain_html_part(self):
        response = post_article_comment(self.data, self.article)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith("/comments/sent/?c="))
        self.assertTrue(self.mock_mailer.call_count == 1)
        self.assertTrue(self.mock_mailer.call_args[1]["html"] is not None)


# ---------------------------------------------------------------------
# Test module level `_*_tmpl` variables. Verify that they include
# the them (settings.COMMENTS_INK_THEME_DIR) in the path when that
# setting is provided.


def test_template_path_includes_theme(monkeypatch):
    monkeypatch.setattr(
        views.settings, "COMMENTS_INK_THEME_DIR", "avatar_in_header"
    )
    importlib.reload(views)
    assert views.theme_dir == "themes/avatar_in_header"
    assert views.theme_dir_exists == True
    monkeypatch.setattr(views.settings, "COMMENTS_INK_THEME_DIR", "")
    importlib.reload(views)


def test_template_path_does_not_include_theme(monkeypatch):
    assert views.theme_dir == ""
    assert views.theme_dir_exists == False


# ---------------------------------------------------------------------
# Test 'post' via accessing the exposed functionality, to
# later replace the implementation with class-based views.


def test_post_view_requires_method_to_be_POST(rf):
    # GET should not work.
    request = rf.get(reverse("comments-ink-post"))
    request._dont_enforce_csrf_checks = True
    response = views.post(request)
    assert response.status_code == 405

    # PUT should not work.
    request = rf.put(reverse("comments-ink-post"))
    request._dont_enforce_csrf_checks = True
    response = views.post(request)
    assert response.status_code == 405

    # PATCH should not work.
    request = rf.patch(reverse("comments-ink-post"))
    request._dont_enforce_csrf_checks = True
    response = views.post(request)
    assert response.status_code == 405


def mocked_post_js(*args, **kwargs):
    return JsonResponse({"post_js_called": True}, status=200)


def test_XMLHttpRequest_post_view_handles_to_post_js(rf, monkeypatch):
    request = rf.post(
        reverse(
            "comments-ink-post",
        ),
        data={},
    )
    request._dont_enforce_csrf_checks = True
    request.META["HTTP_X_REQUESTED_WITH"] = "XMLHttpRequest"
    monkeypatch.setattr(views, "post_js", mocked_post_js)
    response = views.post(request)
    assert response.status_code == 200
    assert json.loads(response.content) == {"post_js_called": True}


# ---------------------------------------------------------------------
class MockedPostBadRequest(CommentPostBadRequest):
    def __init__(self, why):
        self.why = why
        super().__init__(why)


def prepare_comment_form_data(an_article):
    form = django_comments.get_form()(an_article)
    data = {
        "name": "Joe",
        "email": "joe@example.com",
        "followup": True,
        "reply_to": 0,
        "level": 1,
        "order": 1,
        "comment": "Es war einmal eine kleine...",
        "next": reverse("comments-ink-sent"),
    }
    data.update(form.initial)
    return data


def prepare_request_to_post_form(
    monkeypatch, rf, an_article, user=None, remove_fields=[], add_fields=[]
):
    monkeypatch.setattr(views, "CommentPostBadRequest", MockedPostBadRequest)
    data = prepare_comment_form_data(an_article)

    # Remove fields listed in remove_fields.
    for field_name in remove_fields:
        data.pop(field_name)

    # Add fields listed in add_fields as {"name": <name>, "value": <value>}
    for field in add_fields:
        data[field["name"]] = field["value"]

    article_url = reverse(
        "article-detail",
        kwargs={
            "year": an_article.publish.year,
            "month": an_article.publish.month,
            "day": an_article.publish.day,
            "slug": an_article.slug,
        },
    )
    request = rf.post(article_url, data=data, follow=True)
    if user:
        request.user = user
    else:
        request.user = AnonymousUser()
    request._dont_enforce_csrf_checks = True
    return request


@pytest.mark.django_db
def test_post_comment_form_without_content_type(
    monkeypatch, rf, an_article, an_user
):
    request = prepare_request_to_post_form(
        monkeypatch,
        rf,
        an_article,
        user=an_user,
        remove_fields=["content_type"],
    )
    response = views.post(request)
    assert response.status_code == 400
    assert response.why == "Missing content_type or object_pk field."


@pytest.mark.django_db
def test_post_comment_form_without_object_pk(
    monkeypatch, rf, an_article, an_user
):
    request = prepare_request_to_post_form(
        monkeypatch, rf, an_article, user=an_user, remove_fields=["object_pk"]
    )
    response = views.post(request)
    assert response.status_code == 400
    assert response.why == "Missing content_type or object_pk field."


def mock_get_model(etr):
    def _mocked_function(*args, **kwargs):
        raise etr("Something went wrong")

    return _mocked_function


@pytest.mark.django_db
@pytest.mark.parametrize(
    "monkeypatch, rf, an_article, an_user, exc, message",
    [
        (
            "monkeypatch",
            "rf",
            "an_article",
            "an_user",
            LookupError,
            "Invalid content_type value",
        ),
        (
            "monkeypatch",
            "rf",
            "an_article",
            "an_user",
            TypeError,
            "Invalid content_type value",
        ),
        (
            "monkeypatch",
            "rf",
            "an_article",
            "an_user",
            AttributeError,
            "The given content-type",
        ),
        (
            "monkeypatch",
            "rf",
            "an_article",
            "an_user",
            ObjectDoesNotExist,
            "No object matching content-type",
        ),
        (
            "monkeypatch",
            "rf",
            "an_article",
            "an_user",
            ValueError,
            "Attempting to get content-type",
        ),
        (
            "monkeypatch",
            "rf",
            "an_article",
            "an_user",
            ValidationError,
            "Attempting to get content-type",
        ),
    ],
    indirect=["monkeypatch", "rf", "an_article", "an_user"],
)
def test_post_comment_form_raises_an_error(
    monkeypatch, rf, an_article, an_user, exc, message
):
    monkeypatch.setattr(views.apps, "get_model", mock_get_model(exc))
    request = prepare_request_to_post_form(
        monkeypatch, rf, an_article, user=an_user
    )
    response = views.post(request)
    assert response.status_code == 400
    assert response.why.startswith(message)


# ---------------------------------------------------------------------


def get_form_mocked(has_errors=False):
    class MockedForm:
        def __init__(self, target, data=None):
            self.has_errors = has_errors
            self.target = target
            self.data = data

        def security_errors(self):
            return self.has_errors

    return MockedForm


@pytest.mark.django_db
def test_post_comment_form_with_security_errors(
    monkeypatch, rf, an_article, an_user
):
    monkeypatch.setattr(
        views, "get_form", lambda: get_form_mocked(has_errors=True)
    )
    request = prepare_request_to_post_form(
        monkeypatch, rf, an_article, user=an_user
    )
    response = views.post(request)
    assert response.status_code == 400
    assert response.why.startswith("The comment form failed security ")


# ---------------------------------------------------------------------
# Check the preview templates in the following conditions:
#  1st: when the theme_dir is left blank.
#  2nd: when the theme_dir is been given in the settings.
#  3rd: when the form had errors (remove the 'comment' field and add it empty.)


@pytest.mark.django_db
def test_post_comment_form_in_preview_without_theme_dir(
    monkeypatch, rf, an_article, an_user
):
    monkeypatch.setattr(views, "render", lambda x, tmpl_list, y: tmpl_list)
    request = prepare_request_to_post_form(
        monkeypatch,
        rf,
        an_article,
        user=an_user,
        add_fields=[
            {"name": "preview", "value": 1},
        ],
    )
    template_list = views.post(request)
    assert template_list == [
        "comments/tests/article/preview.html",
        "comments/tests/preview.html",
        "comments/preview.html",
    ]


@pytest.mark.django_db
def test_post_comment_form_in_preview_with_theme_dir(
    monkeypatch, rf, an_article, an_user
):
    monkeypatch.setattr(
        views.settings, "COMMENTS_INK_THEME_DIR", "feedback_in_header"
    )
    importlib.reload(views)

    monkeypatch.setattr(views, "render", lambda x, tmpl_list, y: tmpl_list)
    request = prepare_request_to_post_form(
        monkeypatch,
        rf,
        an_article,
        user=an_user,
        add_fields=[
            {"name": "preview", "value": 1},
        ],
    )
    template_list = views.post(request)
    assert template_list == [
        "comments/themes/feedback_in_header/tests/article/preview.html",
        "comments/themes/feedback_in_header/tests/preview.html",
        "comments/themes/feedback_in_header/preview.html",
        "comments/tests/article/preview.html",
        "comments/tests/preview.html",
        "comments/preview.html",
    ]
    # Revert theme.
    monkeypatch.setattr(views.settings, "COMMENTS_INK_THEME_DIR", "")
    importlib.reload(views)


@pytest.mark.django_db
def test_post_comment_form_with_an_empty_comment_field(
    monkeypatch, rf, an_article, an_user
):
    monkeypatch.setattr(views, "render", lambda x, tmpl_list, y: tmpl_list)
    request = prepare_request_to_post_form(
        monkeypatch,
        rf,
        an_article,
        user=an_user,
        remove_fields=["comment"],
        add_fields=[
            {"name": "comment", "value": ""},
        ],
    )
    template_list = views.post(request)
    assert template_list == [
        "comments/tests/article/preview.html",
        "comments/tests/preview.html",
        "comments/preview.html",
    ]


# ----------------------------------------------------------------------
# Test that during the post of the comment form the
# signal 'comment_will_be_posted' has been called.
#  1. Check that when the signal returns False, a HTTP 400 is returned.
#  2. Check that when the signal returns True, the comment is posted.


@pytest.mark.django_db
def test_post_comment_form__comment_will_be_posted__returns_400(
    monkeypatch, rf, an_article, an_user
):
    def mock_send(*args, **kwargs):
        class Receiver:
            def __init__(self):
                self.__name__ = "mocked receiver"

        return [
            (Receiver(), False),
        ]

    monkeypatch.setattr(views.comment_will_be_posted, "send", mock_send)
    request = prepare_request_to_post_form(
        monkeypatch, rf, an_article, user=an_user
    )
    response = views.post(request)
    assert response.status_code == 400
    assert response.why.startswith("comment_will_be_posted receiver")


@pytest.mark.django_db
def test_post_comment_form__comment_will_be_posted__returns_302(
    monkeypatch, rf, an_article, an_user
):
    def mock_send(*args, **kwargs):
        return [
            (None, True),
        ]

    monkeypatch.setattr(views.comment_will_be_posted, "send", mock_send)
    request = prepare_request_to_post_form(
        monkeypatch, rf, an_article, user=an_user
    )
    response = views.post(request)
    assert response.status_code == 302
    assert response.url == "/comments/sent/?c=1"


# ----------------------------------------------------------------------
# Test that 'comment_was_posted' signal has been sent.
# Follow the same approach as here above, but check
# that method 'send' is called.


@pytest.mark.django_db
def test_post_comment_form__comment_was_posted__signal_sent(
    monkeypatch, rf, an_article, an_user
):
    mock_send = Mock()
    monkeypatch.setattr(views.comment_was_posted, "send", mock_send)
    request = prepare_request_to_post_form(
        monkeypatch, rf, an_article, user=an_user
    )
    views.post(request)
    assert mock_send.called


# ---------------------------------------------------------------------
@pytest.mark.django_db
def test_post_comment_form_has_cpage_qs_param(
    monkeypatch, rf, an_article, an_user
):
    request = prepare_request_to_post_form(
        monkeypatch,
        rf,
        an_article,
        user=an_user,
        add_fields=[{"name": "cpage", "value": 2}],
    )
    response = views.post(request)
    assert response.status_code == 302
    assert response.url == "/comments/sent/?c=1&cpage=2"


# ---------------------------------------------------------------------
def prepare_js_request_to_post_form(
    rf, an_article, user=None, remove_fields=[], add_fields=[]
):
    data = prepare_comment_form_data(an_article)

    # Remove fields listed in remove_fields.
    for field_name in remove_fields:
        data.pop(field_name)

    # Add fields listed in add_fields as {"name": <name>, "value": <value>}
    for field in add_fields:
        data[field["name"]] = field["value"]

    article_url = reverse(
        "article-detail",
        kwargs={
            "year": an_article.publish.year,
            "month": an_article.publish.month,
            "day": an_article.publish.day,
            "slug": an_article.slug,
        },
    )
    request = rf.post(article_url, data=data, follow=True)
    if user != None:
        request.user = user
    else:
        request.user = AnonymousUser()
    request._dont_enforce_csrf_checks = True
    request.META["HTTP_X_REQUESTED_WITH"] = "XMLHttpRequest"
    return request


@pytest.mark.django_db
def test_post_js_comment_form_missing_name_and_email(rf, an_article, an_user):
    request = prepare_js_request_to_post_form(
        rf, an_article, user=an_user, remove_fields=["name", "email"]
    )
    response = views.post(request)
    assert response.status_code == 201
    result = json.loads(response.content)
    assert result["html"].find("Your comment has been already published.") > -1
    comment = InkComment.objects.get(pk=1)
    assert comment.user == an_user


@pytest.mark.django_db
@pytest.mark.parametrize(
    "rf, an_article, an_user, remove_field",
    [
        ("rf", "an_article", "an_user", "content_type"),
        ("rf", "an_article", "an_user", "object_pk"),
    ],
    indirect=["rf", "an_article", "an_user"],
)
def test_post_js_comment_form_missing_content_type_or_object_pk(
    rf, an_article, an_user, remove_field
):
    request = prepare_js_request_to_post_form(
        rf, an_article, user=an_user, remove_fields=[remove_field]
    )
    response = views.post(request)
    assert response.status_code == 400
    result = json.loads(response.content)
    assert result["html"].find("Missing content_type or object_pk field.") > -1


@pytest.mark.django_db
@pytest.mark.parametrize(
    "monkeypatch, rf, an_article, an_user, exc, message",
    [
        (
            "monkeypatch",
            "rf",
            "an_article",
            "an_user",
            LookupError,
            "Invalid content_type value",
        ),
        (
            "monkeypatch",
            "rf",
            "an_article",
            "an_user",
            TypeError,
            "Invalid content_type value",
        ),
        (
            "monkeypatch",
            "rf",
            "an_article",
            "an_user",
            AttributeError,
            "The given content-type",
        ),
        (
            "monkeypatch",
            "rf",
            "an_article",
            "an_user",
            ObjectDoesNotExist,
            "No object matching content-type",
        ),
        (
            "monkeypatch",
            "rf",
            "an_article",
            "an_user",
            ValueError,
            "Attempting to get content-type",
        ),
        (
            "monkeypatch",
            "rf",
            "an_article",
            "an_user",
            ValidationError,
            "Attempting to get content-type",
        ),
    ],
    indirect=["monkeypatch", "rf", "an_article", "an_user"],
)
def test_post_js_comment_form_returns_400(
    monkeypatch, rf, an_article, an_user, exc, message
):
    monkeypatch.setattr(views.apps, "get_model", mock_get_model(exc))
    request = prepare_js_request_to_post_form(rf, an_article, user=an_user)
    response = views.post(request)
    assert response.status_code == 400
    result = json.loads(response.content)
    assert result["html"].find(message) > -1


@pytest.mark.django_db
def test_post_js_comment_form_with_security_errors(
    monkeypatch, rf, an_article, an_user
):
    monkeypatch.setattr(
        views, "get_form", lambda: get_form_mocked(has_errors=True)
    )
    request = prepare_js_request_to_post_form(rf, an_article, user=an_user)
    response = views.post(request)
    assert response.status_code == 400
    result = json.loads(response.content)
    assert result["html"].find("The comment form failed security") > -1


# ---------------------------------------------------------------------
# Do a JS Post and check the preview templates in the following
# conditions:
#  1st: when the reply_to field is 0, and the theme_dir is left blank.
#  2nd: when the reply_to field is 1, and the theme_dir is left blank.
#  3nd: when the reply_to is 0 and theme_dir is given in the settings.
#  4th: when the reply_to is 1 and theme_dir is given in the settings.
#  5th: when the form had errors (remove the 'comment' field and add it empty.)


@pytest.mark.django_db
def test_post_js_comment_form_in_preview__no_reply_to__without_theme_dir(
    monkeypatch, rf, an_article, an_user
):
    monkeypatch.setattr(
        views, "json_res", lambda req, tmpl_list, ctx, **kwargs: tmpl_list
    )
    request = prepare_js_request_to_post_form(
        rf,
        an_article,
        user=an_user,
        remove_fields=["reply_to"],
        add_fields=[
            {"name": "preview", "value": 1},
            {"name": "reply_to", "value": 0},  # Will use 'form_js.html'.
        ],
    )
    template_list = views.post(request)
    assert template_list == [
        "comments/tests/article/form_js.html",
        "comments/tests/form_js.html",
        "comments/form_js.html",
    ]


@pytest.mark.django_db
def test_post_js_comment_form_in_preview__with_reply_to__without_theme_dir(
    monkeypatch, rf, an_article, an_user
):
    monkeypatch.setattr(
        views, "json_res", lambda req, tmpl_list, ctx, **kwargs: tmpl_list
    )
    request = prepare_js_request_to_post_form(
        rf,
        an_article,
        user=an_user,
        remove_fields=["reply_to"],
        add_fields=[
            {"name": "preview", "value": 1},
            {"name": "reply_to", "value": 1},  # Will use 'reply_form_js.html'.
        ],
    )
    template_list = views.post(request)
    assert template_list == [
        "comments/tests/article/reply_form_js.html",
        "comments/tests/reply_form_js.html",
        "comments/reply_form_js.html",
    ]


@pytest.mark.django_db
def test_post_js_comment_form_in_preview__no_reply_to__with_theme_dir(
    monkeypatch, rf, an_article, an_user
):
    monkeypatch.setattr(
        views.settings, "COMMENTS_INK_THEME_DIR", "feedback_in_header"
    )
    importlib.reload(views)

    monkeypatch.setattr(
        views, "json_res", lambda req, tmpl_list, ctx, **kwargs: tmpl_list
    )
    request = prepare_js_request_to_post_form(
        rf,
        an_article,
        user=an_user,
        remove_fields=["reply_to"],
        add_fields=[
            {"name": "preview", "value": 1},
            {"name": "reply_to", "value": 0},  # Will use 'form_js.html'.
        ],
    )
    template_list = views.post(request)
    assert template_list == [
        "comments/themes/feedback_in_header/tests/article/form_js.html",
        "comments/themes/feedback_in_header/tests/form_js.html",
        "comments/themes/feedback_in_header/form_js.html",
        "comments/tests/article/form_js.html",
        "comments/tests/form_js.html",
        "comments/form_js.html",
    ]

    # Revert theme.
    monkeypatch.setattr(views.settings, "COMMENTS_INK_THEME_DIR", "")
    importlib.reload(views)


@pytest.mark.django_db
def test_post_js_comment_form_in_preview__with_reply_to__with_theme_dir(
    monkeypatch, rf, an_article, an_user
):
    monkeypatch.setattr(
        views.settings, "COMMENTS_INK_THEME_DIR", "feedback_in_header"
    )
    importlib.reload(views)

    monkeypatch.setattr(
        views, "json_res", lambda req, tmpl_list, ctx, **kwargs: tmpl_list
    )
    request = prepare_js_request_to_post_form(
        rf,
        an_article,
        user=an_user,
        remove_fields=["reply_to"],
        add_fields=[
            {"name": "preview", "value": 1},
            {"name": "reply_to", "value": 1},  # Will use 'reply_form_js.html'.
        ],
    )
    template_list = views.post(request)
    assert template_list == [
        "comments/themes/feedback_in_header/tests/article/reply_form_js.html",
        "comments/themes/feedback_in_header/tests/reply_form_js.html",
        "comments/themes/feedback_in_header/reply_form_js.html",
        "comments/tests/article/reply_form_js.html",
        "comments/tests/reply_form_js.html",
        "comments/reply_form_js.html",
    ]

    # Revert theme.
    monkeypatch.setattr(views.settings, "COMMENTS_INK_THEME_DIR", "")
    importlib.reload(views)


@pytest.mark.django_db
def test_post_js_comment_form_with_an_empty_comment_field(
    rf, an_article, an_user
):
    request = prepare_js_request_to_post_form(
        rf,
        an_article,
        user=an_user,
        remove_fields=["comment"],
        add_fields=[
            {"name": "comment", "value": ""},
        ],
    )
    response = views.post(request)
    assert response.status_code == 200
    result = json.loads(response.content)
    assert result["field_focus"] == "comment"


# ---------------------------------------------------------------------
# Test that during the post_js of the comment form the signal
# 'comment_will_be_posted' has been called.
#  1. Check that when the signal returns False, a HTTP 400 is returned.
#  2. Check that wh nthe signal returns True, the comment is posted.


@pytest.mark.django_db
def test_post_js_comment_form__comment_will_be_posted__returns_400(
    monkeypatch, rf, an_article, an_user
):
    def mock_send(*args, **kwargs):
        class Receiver:
            def __init__(self):
                self.__name__ = "mocked receiver"

        return [
            (Receiver(), False),
        ]

    monkeypatch.setattr(views.comment_will_be_posted, "send", mock_send)
    request = prepare_js_request_to_post_form(rf, an_article, user=an_user)
    response = views.post(request)
    assert response.status_code == 400
    result = json.loads(response.content)
    assert result["html"].find("comment_will_be_posted receiver") > -1


@pytest.mark.django_db
def test_post_js_comment_form__comment_will_be_posted__returns_201(
    monkeypatch, rf, an_article, an_user
):
    def mock_send(*args, **kwargs):
        return [
            (None, True),
        ]

    monkeypatch.setattr(views.comment_will_be_posted, "send", mock_send)
    request = prepare_js_request_to_post_form(rf, an_article, user=an_user)
    response = views.post(request)
    assert response.status_code == 201
    result = json.loads(response.content)
    assert result["html"].find("Your comment has been already published.") > -1
    comment = InkComment.objects.get(pk=1)
    assert comment.user == an_user


@pytest.mark.django_db
def test_sent_js_returns_posted_js_tmpl(monkeypatch, rf, an_article):
    send_mail_mock = Mock()
    monkeypatch.setattr(views.utils, "send_mail", send_mail_mock)
    monkeypatch.setattr(
        views,
        "json_res",
        lambda req, tmpl_list, ctx, status: (tmpl_list, status),
    )
    request = prepare_js_request_to_post_form(rf, an_article)
    template_list, status = views.post(request)
    assert send_mail_mock.called
    assert status == 202
    assert template_list == [
        "comments/tests/article/posted_js.html",
        "comments/tests/posted_js.html",
        "comments/posted_js.html",
    ]


@pytest.mark.django_db
def test_sent_js_failing_to_decode_key_returns_400(monkeypatch, rf, an_article):
    def _raise_exception(*args, **kwargs):
        raise Exception("Something went wrong!")

    send_mail_mock = Mock()
    monkeypatch.setattr(views.signing, "loads", _raise_exception)
    monkeypatch.setattr(views.utils, "send_mail", send_mail_mock)
    request = prepare_js_request_to_post_form(rf, an_article)
    response = views.post(request)
    assert send_mail_mock.called
    result = json.loads(response.content)
    assert response.status_code == 400
    assert result["html"].find("Comment does not exist") > -1


@pytest.mark.django_db
def test_sent_js__returns_moderated_js_tmpl(
    monkeypatch, rf, an_article, an_user
):
    def mock_send(*args, **kwargs):
        comment = kwargs["comment"]
        comment.is_public = False
        comment.save()
        return [
            (None, True),
        ]

    monkeypatch.setattr(views.comment_will_be_posted, "send", mock_send)

    monkeypatch.setattr(
        views,
        "json_res",
        lambda req, tmpl_list, ctx, status: (tmpl_list, status),
    )

    request = prepare_js_request_to_post_form(rf, an_article, user=an_user)
    template_list, status = views.post(request)
    assert status == 201
    assert template_list == [
        "comments/tests/article/moderated_js.html",
        "comments/tests/moderated_js.html",
        "comments/moderated_js.html",
    ]
    comment = InkComment.objects.get(pk=1)
    assert comment.user == an_user
    assert comment.is_public == False


# ---------------------------------------------------------------------
# Test the redirect to 'sent'.


@pytest.mark.django_db
def test_sent_receives_non_existing_comment_pk(monkeypatch, rf):
    monkeypatch.setattr(views, "CommentPostBadRequest", MockedPostBadRequest)
    request = rf.get("/comments/sent/")
    response = views.sent(request)
    assert response.status_code == 400
    assert response.why == "Comment doesn't exist"


@pytest.mark.django_db
def test_redirect_to_sent__auth_comment__returns_comments_url(
    monkeypatch, rf, an_article, an_user
):
    request = prepare_request_to_post_form(
        monkeypatch, rf, an_article, user=an_user
    )
    response = views.post(request)
    assert response.status_code == 302
    assert response.url == "/comments/sent/?c=1"

    request = rf.get(response.url)
    request.user = an_user
    response = views.sent(request)
    assert response.status_code == 302
    comment = InkComment.objects.get(pk=1)
    assert response.url == comment.get_absolute_url()


@pytest.mark.django_db
def test_redirect_to_sent__not_auth_comment__returns_posted_tmpl(
    monkeypatch, rf, an_article
):
    monkeypatch.setattr(
        views.utils, "send_mail", lambda *args, **kwargs: Mock(*args, *kwargs)
    )
    request = prepare_request_to_post_form(monkeypatch, rf, an_article)
    response = views.post(request)
    assert response.status_code == 302
    assert response.url.startswith("/comments/sent/?c=")
    request = rf.get(response.url)
    response = views.sent(request)
    assert response.status_code == 200
    assert (
        response.content.decode("utf-8").find("Comment confirmation requested")
        > -1
    )


@pytest.mark.django_db
def test_redirect_to_sent__not_auth_comment__uses_posted_tmpl(
    monkeypatch, rf, an_article
):
    monkeypatch.setattr(views, "render", lambda x, tmpl_list, y: tmpl_list)
    request = prepare_request_to_post_form(monkeypatch, rf, an_article)
    response = views.post(request)
    assert response.status_code == 302
    assert response.url.startswith("/comments/sent/?c=")
    request = rf.get(response.url)
    template = views.sent(request)
    assert template == "comments/posted.html"


@pytest.mark.django_db
def test_redirect_to_sent__bad_key__returns_400(monkeypatch, rf, an_article):
    request = prepare_request_to_post_form(monkeypatch, rf, an_article)
    response = views.post(request)
    assert response.status_code == 302
    assert response.url.startswith("/comments/sent/?c=")
    request = rf.get(response.url[:-1])  # Force malformed key.
    response = views.sent(request)
    assert response.status_code == 400
    assert response.why == "Comment doesn't exist"


@pytest.mark.django_db
def test_redirect_to_sent__not_public_comment__comment_in_moderation(
    monkeypatch, rf, an_article, an_user
):
    def create_comment(tmp_comment):
        tmp_comment.pop("page_number", None)
        comment = InkComment(**tmp_comment)
        comment.is_public = False  # To force rendering moderated_tmpl.
        comment.save()
        return comment

    monkeypatch.setattr(views, "_create_comment", create_comment)
    request = prepare_request_to_post_form(
        monkeypatch, rf, an_article, user=an_user
    )
    response = views.post(request)
    assert response.status_code == 302
    assert response.url.startswith("/comments/sent/?c=")
    request = rf.get(response.url)
    response = views.sent(request)
    assert response.status_code == 200
    assert (
        response.content.decode("utf-8").find("Your comment is in moderation.")
        > -1
    )


@pytest.mark.django_db
def test_redirect_to_sent__not_public_comment__uses_moderated_tmpl(
    monkeypatch, rf, an_article, an_user
):
    def create_comment(tmp_comment):
        tmp_comment.pop("page_number", None)
        comment = InkComment(**tmp_comment)
        comment.is_public = False  # To force rendering moderated_tmpl.
        comment.save()
        return comment

    monkeypatch.setattr(
        views.utils, "send_mail", lambda *args, **kwargs: Mock(*args, *kwargs)
    )
    monkeypatch.setattr(views, "_create_comment", create_comment)
    monkeypatch.setattr(views, "render", lambda x, tmpl_list, y: tmpl_list)
    request = prepare_request_to_post_form(
        monkeypatch, rf, an_article, user=an_user
    )
    response = views.post(request)
    assert response.status_code == 302
    assert response.url.startswith("/comments/sent/?c=")
    request = rf.get(response.url)
    template_list = views.sent(request)
    assert template_list == [
        "comments/tests/article/moderated.html",
        "comments/tests/moderated.html",
        "comments/moderated.html",
    ]


# ---------------------------------------------------------------------
@pytest.mark.django_db
def test_GET_react_without_reactions_enabled_raises(
    monkeypatch, rf, an_user, an_articles_comment
):
    def raise_PermissionDenied(*args):
        raise PermissionDenied(detail="Mee", code=status.HTTP_403_FORBIDDEN)

    monkeypatch.setattr(views.utils, "check_option", raise_PermissionDenied)
    request = rf.get(
        reverse("comments-ink-react", args=(an_articles_comment.pk,))
    )
    request.user = an_user
    # request._dont_enforce_csrf_checks = True
    try:
        views.react(request, an_articles_comment.pk)
    except Exception as exc:
        assert exc.detail == ErrorDetail(string="Mee", code=403)


@pytest.mark.django_db
def test_GET_react_renders_react_tmpl(
    monkeypatch, rf, an_user, an_articles_comment
):
    monkeypatch.setattr(views.utils, "check_option", lambda *_: True)
    monkeypatch.setattr(views, "render", lambda x, tmpl, ctx: (tmpl, ctx))
    request = rf.get(
        reverse("comments-ink-react", args=(an_articles_comment.pk,))
    )
    request.user = an_user
    template, context = views.react(request, an_articles_comment.pk)
    assert template == "comments/react.html"
    ctx_keys = [
        "comment",
        "user_reactions",
        "next",
        "page_number",
        "comments_page_qs_param",
        "comments_fold_qs_param",
    ]
    for key in ctx_keys:
        assert key in context


@pytest.mark.django_db
def test_POST_react(monkeypatch, rf, an_user, an_articles_comment):
    monkeypatch.setattr(views.utils, "check_option", lambda *_: True)
    monkeypatch.setattr(views, "render", lambda x, tmpl, ctx: (tmpl, ctx))
    request = rf.post(
        reverse("comments-ink-react", args=(an_articles_comment.pk,)),
        data={"reaction": "+"},
        follow=True,
    )
    request.user = an_user
    request._dont_enforce_csrf_checks = True
    response = views.react(request, an_articles_comment.pk)
    assert response.status_code == 302
    request = rf.get(response.url)
    template, context = views.react_done(request)
    assert template == "comments/reacted.html"
    assert context["comment"] == an_articles_comment
    assert context["cpage"] == 1
    assert context["cfold"] == ""


@pytest.mark.django_db
def test_POST_react_two_users_add_same_reaction_2nd_user_withdraws_it(
    monkeypatch, rf, an_user, an_user_2, an_articles_comment
):
    # This function tests perform_react: two users add the same
    # reaction to the same comment, and then the second user withdraws the
    # reaction.
    def get_cr_counter():
        return CommentReaction.objects.get(
            reaction="+", comment=an_articles_comment
        ).counter

    def prepare_request(user):
        request = rf.post(
            reverse("comments-ink-react", args=(an_articles_comment.pk,)),
            data={"reaction": "+"},
            follow=True,
        )
        request.user = user
        request._dont_enforce_csrf_checks = True
        return request

    monkeypatch.setattr(views.utils, "check_option", lambda *_: True)

    # 1: an_user sends the "+" reaction to this comment.
    request = prepare_request(an_user)
    response = views.react(request, an_articles_comment.pk)
    assert response.status_code == 302
    request = rf.get(response.url)
    response = views.react_done(request)
    assert response.status_code == 200
    assert get_cr_counter() == 1

    # 2: an_user_2 sends the "+" reaction to this comment.
    request = prepare_request(an_user_2)
    response = views.react(request, an_articles_comment.pk)
    assert response.status_code == 302
    request = rf.get(response.url)
    response = views.react_done(request)
    assert response.status_code == 200
    assert get_cr_counter() == 2

    # 3: an_user_2 sends again the "+" reaction to this comment,
    # the effect is the withdrawal of the reaction.
    request = prepare_request(an_user_2)
    response = views.react(request, an_articles_comment.pk)
    assert response.status_code == 302
    request = rf.get(response.url)
    response = views.react_done(request)
    assert response.status_code == 200
    assert get_cr_counter() == 1

    # 4: an_user sends again the "+" reaction to this comment,
    # the effect is the deletion of the CommentReaction (as it's the last
    # author of the CommentReaction).
    request = prepare_request(an_user)
    response = views.react(request, an_articles_comment.pk)
    assert response.status_code == 302
    request = rf.get(response.url)
    response = views.react_done(request)
    assert response.status_code == 200
    with pytest.raises(CommentReaction.DoesNotExist):
        get_cr_counter()


@pytest.mark.django_db
def test_POST_react_js(monkeypatch, rf, an_user, an_articles_comment):
    monkeypatch.setattr(views.utils, "check_option", lambda *_: True)
    monkeypatch.setattr(
        views,
        "json_res",
        lambda req, tmpl_list, ctx, **kwargs: (
            tmpl_list,
            ctx,
            kwargs["status"],
        ),
    )
    request = rf.post(
        reverse("comments-ink-react", args=(an_articles_comment.pk,)),
        data={"reaction": "+"},
        follow=True,
    )
    request.user = an_user
    request._dont_enforce_csrf_checks = True
    request.META["HTTP_X_REQUESTED_WITH"] = "XMLHttpRequest"
    tmpl_list, context, status = views.react(request, an_articles_comment.pk)
    assert status == 201
    assert tmpl_list == [
        "comments/tests/article/comment_reactions.html",
        "comments/tests/comment_reactions.html",
        "comments/comment_reactions.html",
    ]
    assert "comment" in context
    assert context["comment"] == an_articles_comment


@pytest.mark.django_db
def test_GET_react_with_an_existing_comments_reaction(
    monkeypatch, rf, an_user, an_articles_comment, a_comments_reaction
):
    # fixture a_comments_reaction is needed in the signature of the test,
    # so that it creates a_comments_reaction. When we then request GET the
    # react view we will get that reaction with the 'active' css class.

    monkeypatch.setattr(views.utils, "check_option", lambda *_: True)
    request = rf.get(
        reverse("comments-ink-react", args=(an_articles_comment.pk,)),
    )
    request.user = an_user
    response = views.react(request, an_articles_comment.pk)
    assert response.status_code == 200
    html = response.content.replace(b"  ", b" ").replace(b"\n", b"")
    re_str = (  # This regexp matches the class="secondary active" in HTML.
        rb'<button\s+type="submit"\s+name="reaction"\s+value="\+"'
        rb'\s+class="(?P<classes>[a-z\s]+)"\s*>'
    )
    css_classes = re.search(re_str, html).group("classes")
    assert css_classes.find(b"active") > -1


@pytest.mark.django_db
def test_react_done_raises_404_if_no_param_c_or_comment_does_not_exist(rf):
    request = rf.get(reverse("comments-ink-react-done"))
    with pytest.raises(Http404):
        views.react_done(request)

    request = rf.get(reverse("comments-ink-react-done") + "?c=1")
    with pytest.raises(Http404):
        views.react_done(request)


# ---------------------------------------------------------------------
# Test view's get_inkcomment_url.


@pytest.mark.django_db
def test_get_inkcomment_url_without_page(rf, an_article, an_articles_comment):
    args = (
        an_articles_comment.content_type.pk,
        int(an_articles_comment.object_pk),
        an_articles_comment.pk,
    )
    request = rf.get(reverse("comments-url-redirect", args=args))
    response = views.get_inkcomment_url(request, *args)
    assert response.status_code == 302
    assert response.url.endswith(an_article.get_absolute_url())


@pytest.mark.django_db
def test_get_inkcomment_url_with_bad_page(rf, an_articles_comment):
    args = (
        an_articles_comment.content_type.pk,
        int(an_articles_comment.object_pk),
        an_articles_comment.pk,
    )
    request = rf.get(
        reverse("comments-url-redirect", args=args), {"cpage": "bad"}
    )
    with pytest.raises(Http404):
        views.get_inkcomment_url(request, *args)


@pytest.mark.django_db
@pytest.mark.parametrize(
    "rf, an_article, an_articles_comment, cpage",
    [
        ("rf", "an_article", "an_articles_comment", "last"),
        ("rf", "an_article", "an_articles_comment", "2"),
    ],
    indirect=["rf", "an_article", "an_articles_comment"],
)
def test_get_inkcomment_url_with_page(
    rf, an_article, an_articles_comment, cpage
):
    args = (
        an_articles_comment.content_type.pk,
        int(an_articles_comment.object_pk),
        an_articles_comment.pk,
    )
    request = rf.get(
        reverse("comments-url-redirect", args=args), {"cpage": cpage}
    )
    response = views.get_inkcomment_url(request, *args)
    assert response.status_code == 302
    assert response.url.endswith(
        an_article.get_absolute_url() + f"?cpage={cpage}"
    )



from unittest.mock import patch

import django_comments
from django.test import TestCase
from django_comments_ink.forms import InkCommentForm
from django_comments_ink.models import TmpInkComment
from django_comments_ink.tests.models import Article


class GetFormTestCase(TestCase):
    def test_get_form(self):
        # check function django_comments_ink.get_form retrieves the due class
        self.assertTrue(django_comments.get_form() == InkCommentForm)


class InkCommentFormTestCase(TestCase):
    def setUp(self):
        self.article = Article.objects.create(
            title="September",
            slug="september",
            body="What I did on September...",
        )
        self.form = django_comments.get_form()(self.article)

    def test_get_comment_model(self):
        # check get_comment_model retrieves the due model class
        self.assertTrue(self.form.get_comment_model() == TmpInkComment)

    def test_get_comment_create_data(self):
        # as it's used in django_comments.views.comments
        data = {
            "name": "Daniela",
            "email": "danirus@eml.cc",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal iene kleine...",
        }
        data.update(self.form.initial)
        form = django_comments.get_form()(self.article, data)
        self.assertTrue(self.form.security_errors() == {})
        self.assertTrue(self.form.errors == {})
        comment = form.get_comment_object()

        # it does have the new field 'followup'
        self.assertTrue("followup" in comment)

    @patch.multiple(
        "django_comments_ink.conf.settings", COMMENTS_INK_DEFAULT_FOLLOWUP=False
    )
    def test_followup_prechecked(self):
        # Have to update form object to re-initialize 'followup' checkbox
        # with settings.
        self.form = django_comments.get_form()(self.article)
        self.assertEqual(self.form.fields["followup"].initial, False)

    @patch.multiple(
        "django_comments_ink.conf.settings", COMMENTS_INK_DEFAULT_FOLLOWUP=True
    )
    def test_followup_preunchecked(self):
        # Have to update form object to re-initialize 'followup' checkbox
        # with settings.
        self.form = django_comments.get_form()(self.article)
        self.assertEqual(self.form.fields["followup"].initial, True)



import hashlib
import json
import logging
import re
from urllib.parse import urlencode

from django.apps import apps
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.paginator import InvalidPage, PageNotAnInteger
from django.db.models import Q
from django.http import Http404
from django.template import (
    Library,
    Node,
    TemplateSyntaxError,
    Variable,
    VariableDoesNotExist,
    loader,
)
from django.urls import reverse
from django.utils.module_loading import import_string
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _
from django.views.decorators.csrf import csrf_protect
from django_comments.templatetags.comments import (
    BaseCommentNode,
    RenderCommentFormNode,
)
from django_comments_ink import (
    caching,
    get_comment_reactions_enum,
    get_model,
    get_object_reactions_enum,
    utils,
)
from django_comments_ink.api import frontend
from django_comments_ink.conf import settings
from django_comments_ink.models import (
    ObjectReaction,
    max_thread_level_for_content_type,
)
from django_comments_ink.paginator import CommentsPaginator

logger = logging.getLogger(__name__)

register = Library()


if len(settings.COMMENTS_INK_THEME_DIR) > 0:
    theme_dir = "themes/%s" % settings.COMMENTS_INK_THEME_DIR
    theme_dir_exists = utils.does_theme_dir_exist(f"comments/{theme_dir}")
else:
    theme_dir = ""
    theme_dir_exists = False


# List of possible paths to the list.html template file.
_list_html_tmpl = []
if theme_dir_exists:
    _list_html_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/list.html",
            "comments/{theme_dir}/{app_label}/list.html",
            "comments/{theme_dir}/list.html",
        ]
    )
_list_html_tmpl.extend(
    [
        "comments/{app_label}/{model}/list.html",
        "comments/{app_label}/list.html",
        "comments/list.html",
    ]
)


# List of possible paths to the form.html template file.
_form_html_tmpl = []
if theme_dir_exists:
    _form_html_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/form.html",
            "comments/{theme_dir}/{app_label}/form.html",
            "comments/{theme_dir}/form.html",
        ]
    )
_form_html_tmpl.extend(
    [
        "comments/{app_label}/{model}/form.html",
        "comments/{app_label}/form.html",
        "comments/form.html",
    ]
)


# List of possible paths to the reply_template.html template file.
_reply_template_html_tmpl = []
if theme_dir_exists:
    _reply_template_html_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/reply_template.html",
            "comments/{theme_dir}/{app_label}/reply_template.html",
            "comments/{theme_dir}/reply_template.html",
        ]
    )
_reply_template_html_tmpl.extend(
    [
        "comments/{app_label}/{model}/reply_template.html",
        "comments/{app_label}/reply_template.html",
        "comments/reply_template.html",
    ]
)


_reactions_panel_template_tmpl = []
if theme_dir_exists:
    _reactions_panel_template_tmpl.extend(
        [
            "comments/{theme_dir}/reactions_panel_template.html",
            "comments/reactions_panel_template.html",
        ]
    )
_reactions_panel_template_tmpl.append("comments/reactions_panel_template.html")


_reactions_buttons_tmpl = []
if theme_dir_exists:
    _reactions_buttons_tmpl.extend(
        [
            "comments/{theme_dir}/reactions_buttons.html",
            "comments/reactions_buttons.html",
        ]
    )
_reactions_buttons_tmpl.append("comments/reactions_buttons.html")


# List of possible paths to the object_reactions.html template file.
_object_reactions_tmpl = []
if theme_dir_exists:
    _object_reactions_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/object_reactions.html"
            "comments/{theme_dir}/{app_label}/object_reactions_form.html",
            "comments/{theme_dir}/object_reactions_form.html",
        ]
    )
_object_reactions_tmpl.extend(
    [
        "comments/{app_label}/{model}/object_reactions.html",
        "comments/{app_label}/object_reactions.html",
        "comments/object_reactions.html",
    ]
)


# List of possible paths to the object_reactions_form.html template file.
_object_reactions_form_tmpl = []
if theme_dir_exists:
    _object_reactions_form_tmpl.extend(
        [
            (
                "comments/{theme_dir}/{app_label}/{model}/"
                "object_reactions_form.html"
            ),
            "comments/{theme_dir}/{app_label}/object_reactions_form.html",
            "comments/{theme_dir}/object_reactions_form.html",
        ]
    )
_object_reactions_form_tmpl.extend(
    [
        "comments/{app_label}/{model}/object_reactions_form.html",
        "comments/{app_label}/object_reactions_form.html",
        "comments/object_reactions_form.html",
    ]
)


def filter_folded_comments(qs, context):
    request = context.get("request", None)
    cfold_qs_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM
    cfold = (request and request.GET.get(cfold_qs_param, None)) or None
    if cfold:
        cfold_list = {int(cid) for cid in cfold.split(",")}
        return qs.filter(~Q(level__gt=0, thread_id__in=cfold_list))
    return qs


def folded_comments(context):
    """
    Returns dict with folded comments data to the given context.
    """
    request = context.get("request", None)
    cfold_qs_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM
    cfold = (request and request.GET.get(cfold_qs_param, None)) or None
    fold = (cfold and {int(cid) for cid in cfold.split(",")}) or {}

    return {
        "comments_fold_qs_param": cfold_qs_param,
        cfold_qs_param: ",".join([str(cid) for cid in fold]),
    }


def paginate_queryset(queryset, context, ckey_prefix):
    """
    Returns dict with pagination data for the given queryset and context.
    """
    request = context.get("request", None)
    num_orphans = settings.COMMENTS_INK_MAX_LAST_PAGE_ORPHANS
    page_size = settings.COMMENTS_INK_ITEMS_PER_PAGE

    cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
    page = (request and request.GET.get(cpage_qs_param, None)) or 1

    cfold_qs_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM
    cfold = (request and request.GET.get(cfold_qs_param, None)) or None
    fold = (cfold and {int(cid) for cid in cfold.split(",")}) or {}

    if page_size == 0:
        return {
            "paginator": None,
            "page_obj": None,
            "is_paginated": False,
            "comments_page_qs_param": cpage_qs_param,
            "comment_list": queryset,
            cpage_qs_param: 1,
        }

    paginator = CommentsPaginator(
        queryset,
        page_size,
        orphans=num_orphans,
        comments_folded=fold,
        cache_key_prefix=ckey_prefix,
    )

    try:
        page_number = int(page)
    except ValueError:
        if page == "last":
            page_number = paginator.num_pages
        else:
            raise Http404(
                _("Page is not “last”, nor can it " "be converted to an int.")
            )

    try:
        page = paginator.page(page_number)
        return {
            "paginator": paginator,
            "page_obj": page,
            "is_paginated": page.has_other_pages(),
            "comments_page_qs_param": cpage_qs_param,
            "comment_list": page.object_list,
            cpage_qs_param: page.number,
        }
    except InvalidPage as exc:
        raise Http404(
            _("Invalid page (%(page_number)s): %(message)s")
            % {"page_number": page_number, "message": str(exc)}
        )


class BaseInkCommentNode(BaseCommentNode):
    def get_queryset(self, context):
        ctype, object_pk = self.get_target_ctype_pk(context)
        if not object_pk:
            return self.comment_model.objects.none()

        site_id = utils.get_current_site_id(context.get("request", None))

        kwargs = {
            "ctype_pk": ctype.pk,
            "object_pk": object_pk,
            "site_id": site_id,
        }
        comments_qs_ptn = settings.COMMENTS_INK_CACHE_KEYS["comments_qs"]
        comments_count_ptn = settings.COMMENTS_INK_CACHE_KEYS["comments_count"]
        comments_paged_ptn = settings.COMMENTS_INK_CACHE_KEYS["comments_paged"]
        self.ckey_comments_qs = comments_qs_ptn.format(**kwargs)
        self.ckey_comments_count = comments_count_ptn.format(**kwargs)
        self.ckey_comments_paged = comments_paged_ptn.format(**kwargs)

        # Check whether there is already a qs in the dci cache.
        qs = None
        dci_cache = caching.get_cache()
        if dci_cache != None and self.ckey_comments_qs != "":
            cached = dci_cache.get(self.ckey_comments_qs)
            if cached:
                logger.debug("Get %s from the cache", self.ckey_comments_qs)
                qs = cached

        if not qs:
            qs = super().get_queryset(context)
            if dci_cache != None and self.ckey_comments_qs != "":
                logger.debug("Adding %s to the cache", self.ckey_comments_qs)
                dci_cache.set(self.ckey_comments_qs, qs, timeout=None)

        return qs


class InkCommentListNode(BaseInkCommentNode):
    def get_context_value_from_queryset(self, context, qs):
        return qs


class RenderInkCommentListNode(InkCommentListNode):
    """Render the comment list directly."""

    @classmethod
    def handle_token(cls, parser, token):
        """Class method to parse render_comment_list and return a Node."""
        tokens = token.split_contents()
        if tokens[1] != "for":
            raise TemplateSyntaxError(
                "Second argument in %r tag must be 'for'" % tokens[0]
            )

        # {% render_comment_list for obj %}
        if len(tokens) == 3:
            return cls(object_expr=parser.compile_filter(tokens[2]))

        # {% render_comment_list for app.models pk %}
        elif len(tokens) == 4:
            return cls(
                ctype=BaseCommentNode.lookup_content_type(tokens[2], tokens[0]),
                object_pk_expr=parser.compile_filter(tokens[3]),
            )

    def render(self, context):
        try:
            ctype, object_pk = self.get_target_ctype_pk(context)
        except AttributeError:
            # in get_target_ctype_pk the call to FilterExpression.resolve does
            # not raise VariableDoesNotExist, however in a latter step an
            # AttributeError is raised when the object_expr does not exist
            # in the context. Therefore, this except raises when used as:
            # {% render_inkcomment_list for var_not_in_context %}
            return ""

        template_list = [
            pth.format(
                theme_dir=theme_dir,
                app_label=ctype.app_label,
                model=ctype.model,
            )
            for pth in _list_html_tmpl
        ]

        qs = self.get_queryset(context)
        qs = self.get_context_value_from_queryset(context, qs)
        context_dict = context.flatten()
        qs = filter_folded_comments(qs, context)
        context_dict.update(folded_comments(context))
        context_dict.update(
            paginate_queryset(qs, context, self.ckey_comments_paged)
        )

        # Pass max_thread_level in the context.
        app_model = "%s.%s" % (ctype.app_label, ctype.model)
        MTL = settings.COMMENTS_INK_MAX_THREAD_LEVEL_BY_APP_MODEL
        mtl = MTL.get(app_model, settings.COMMENTS_INK_MAX_THREAD_LEVEL)
        context_dict.update(
            {
                "dcx_theme_dir": theme_dir,
                "max_thread_level": mtl,
                "reply_stack": [],  # List to control reply widget rendering.
            }
        )

        # get_app_model_options returns a dict like: {
        #     'who_can_post': 'all' | 'users',
        #     'check_input_allowed': 'string path to function',
        #     'comment_flagging_enabled': <boolean>,
        #     'comment_reactions_enabled': <boolean>,
        #     'object_reactions_enabled': <boolean>
        # }
        options = utils.get_app_model_options(content_type=app_model)
        check_input_allowed_str = options.pop("check_input_allowed")
        check_func = import_string(check_input_allowed_str)
        target_obj = ctype.get_object_for_this_type(pk=object_pk)

        # Call the function that checks whether comments input
        # is still allowed on the given target_object. And add
        # the resulting boolean to the template context.
        #
        options["is_input_allowed"] = check_func(target_obj)
        context_dict.update(options)

        liststr = loader.render_to_string(template_list, context_dict)
        return liststr


@register.tag
def render_inkcomment_list(parser, token):
    """
    Render the comment list (as returned by ``{% get_inkcomment_list %}``)
    through the ``comments/list.html`` templates.

    Syntax::

        {% render_inkcomment_list for [object] %}
        {% render_inkcomment_list for [app].[model] [obj_id] %}

    Example usage::

        {% render_inkcomment_list for post %}

    """
    return RenderInkCommentListNode.handle_token(parser, token)


# ---------------------------------------------------------------------
class InkCommentCountNode(BaseInkCommentNode):
    """Insert a count of comments into the context."""

    def get_context_value_from_queryset(self, context, qs):
        return qs.count()

    def render(self, context):
        ctype, object_pk = self.get_target_ctype_pk(context)
        if not object_pk:
            context[self.as_varname] = 0
        site_id = utils.get_current_site_id(context.get("request", None))

        result = None
        dci_cache = caching.get_cache()
        key = settings.COMMENTS_INK_CACHE_KEYS["comments_count"].format(
            ctype_pk=ctype.pk, object_pk=object_pk, site_id=site_id
        )

        if dci_cache != None and key != "":
            cached = dci_cache.get(key)
            if cached:
                logger.debug("Fetching %s from the cache", key)
                result = cached

        if not result:
            qs = self.get_queryset(context)
            result = self.get_context_value_from_queryset(context, qs)
            if dci_cache != None and key != "":
                logger.debug("Adding %s to the cache", key)
                dci_cache.set(key, result, timeout=None)

        context[self.as_varname] = result
        return ""


@register.tag
def get_inkcomment_count(parser, token):
    """
    Gets the comment count for the given params and populates the template
    context with a variable containing that value, whose name is defined by the
    'as' clause.

    Syntax::

        {% get_comment_count for [object] as [varname]  %}
        {% get_comment_count for [app].[model] [object_id] as [varname]  %}

    Example usage::

        {% get_comment_count for event as comment_count %}
        {% get_comment_count for calendar.event event.id as comment_count %}
        {% get_comment_count for calendar.event 17 as comment_count %}

    """
    return InkCommentCountNode.handle_token(parser, token)


# ---------------------------------------------------------------------
class RenderInkCommentFormNode(RenderCommentFormNode):
    """
    Almost identical to django_comments' RenderCommentFormNode.

    This class rewrites the `render` method of its parent class, to
    fetch the template from the theme directory.
    """

    def render(self, context):
        try:
            ctype, _ = self.get_target_ctype_pk(context)
            template_list = [
                pth.format(
                    theme_dir=theme_dir,
                    app_label=ctype.app_label,
                    model=ctype.model,
                )
                for pth in _form_html_tmpl
            ]
            context_dict = context.flatten()
            context_dict["form"] = self.get_form(context)
            formstr = loader.render_to_string(template_list, context_dict)
            return formstr
        except:
            return ""


@register.tag
def render_inkcomment_form(parser, token):
    """
    Render "comments/<theme_dir>/form.html" or "comments/form.html".

    Syntax::

        {% get_inkcomment_form for [object] as [varname] %}
        {% get_inkcomment_form for [app].[model] [object_id] as [varname] %}
    """
    return RenderInkCommentFormNode.handle_token(parser, token)


class RenderCommentReplyTemplateNode(RenderCommentFormNode):
    def render(self, context):
        cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
        try:
            ctype, _ = self.get_target_ctype_pk(context)
            template_list = [
                pth.format(
                    theme_dir=theme_dir,
                    app_label=ctype.app_label,
                    model=ctype.model,
                )
                for pth in _reply_template_html_tmpl
            ]
            context_dict = context.flatten()
            context_dict.update(
                {
                    "form": self.get_form(context),
                    "comments_page_qs_param": cpage_qs_param,
                    "dcx_theme_dir": theme_dir,
                }
            )
            formstr = loader.render_to_string(template_list, context_dict)
            return formstr
        except:
            return ""


@register.tag
def render_comment_reply_template(parser, token):
    """
    Render the comment reply form to be used by the JavaScript plugin.

    Syntax::

        {% render_comment_reply_template for [object] %}
        {% render_comment_reply_template for [app].[model] [object_id] %}
    """
    return RenderCommentReplyTemplateNode.handle_token(parser, token)


# ---------------------------------------------------------------------
class RenderQSParams(Node):
    def __init__(self, page_expr, comment_action, comment_object):
        self.page_expr = None
        self.comment_action = comment_action
        self.comment_object = None
        if page_expr != None:
            self.page_expr = Variable(page_expr)
        if comment_object != None:
            self.comment_object = Variable(comment_object)

    def render(self, context):
        cobj = None
        qs_params = []
        cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
        cfold_qs_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM
        fold_param = context.get(cfold_qs_param, None) or ""

        if self.page_expr:
            page = self.page_expr.resolve(context)
            if page != 1:
                qs_params.append(f"{cpage_qs_param}={page}")

        if self.comment_object:
            cobj = self.comment_object.resolve(context)

        if self.comment_object == None and self.page_expr == None:
            page = context.get(cpage_qs_param, None) or None
            if page != 1:
                qs_params.append(f"{cpage_qs_param}={page}")

        if cobj:
            if len(fold_param) > 0:
                fold = {int(cid) for cid in fold_param.split(",")}
            else:
                fold = set()

            if self.comment_action == "fold":
                fold.add(cobj.id)
            elif self.comment_action == "unfold":
                fold.remove(cobj.id)

            fold_csv = ",".join(sorted([str(cid) for cid in fold]))

            if self.page_expr == None:
                request = context.get("request", None)
                page = utils.get_comment_page_number(
                    request,
                    cobj.content_type.id,
                    cobj.object_pk,
                    cobj.id,
                    comments_folded=fold,
                )
                qs_params.append(f"{cpage_qs_param}={page}")

            if len(fold_csv):
                qs_params.append(f"{cfold_qs_param}={fold_csv}")

        else:  # cobj is None.
            if len(fold_param) > 0:
                qs_params.append(f"{cfold_qs_param}={fold_param}")

        return mark_safe(f"{'&'.join(qs_params)}")


@register.tag
def render_qs_params(parser, token):
    """
    Render_qs_params to manage comments page and comments muted list.

    Syntax::

        {% render_qs_params [page <var>] [fold|unfold <InkComment>] %}
    """
    bits = token.contents.split()
    tag_name, args = bits[0], bits[1:]

    if len(args) != 0 and len(args) != 2 and len(args) != 4:
        raise TemplateSyntaxError(
            "%r tag requires either 0 or 2 or 4 arguments" % tag_name
        )

    page_expr = None
    comment_action = None
    comment_object = None

    if len(args) == 2:
        if args[0] == "page":
            page_expr = args[1]
        elif args[0] in ["fold", "unfold"]:
            comment_action = args[0]
            comment_object = args[1]
    elif len(args) == 4:
        if args[0] == "page" and args[2] in ["fold", "unfold"]:
            page_expr = args[1]
            comment_action = args[2]
            comment_object = args[3]
        elif args[0] in ["fold", "unfold"] and args[2] == "page":
            page_expr = args[3]
            comment_action = args[0]
            comment_object = args[1]
        else:
            raise TemplateSyntaxError(
                "%r tag requires 'page <page_var>' and/or "
                "'fold|unfold <comment>' pair arguments" % tag_name
            )

    return RenderQSParams(page_expr, comment_action, comment_object)


@register.filter
def has_comment(cfold, comment):
    if cfold and len(cfold) > 0:
        fold = {int(cid) for cid in cfold.split(",")}
        return comment.id in fold
    return False


@register.filter
def get_anchor(comment, anchor_pattern=None):
    if anchor_pattern:
        cm_abs_url = comment.get_absolute_url(anchor_pattern)
    else:
        cm_abs_url = comment.get_absolute_url()

    hash_pos = cm_abs_url.find("#")
    return cm_abs_url[hash_pos + 1 :]


# ---------------------------------------------------------------------
@register.simple_tag()
def get_inkcomment_permalink(
    comment, page_number=None, comments_folded=None, anchor_pattern=None
):
    """
    Get the permalink for a comment, optionally specifying the format of the
    named anchor to be appended to the end of the URL.

    Usage::
        {% get_inkcomment_permalink comment <page_num> <comments_folded> <anchor_pattern> %}

    Example::
        {% get_inkcomment_permalink comment 2 "1,97,145" "#c%(id)s" %}
    """
    try:
        if anchor_pattern:
            cm_abs_url = comment.get_absolute_url(anchor_pattern)
        else:
            cm_abs_url = comment.get_absolute_url()

        hash_pos = cm_abs_url.find("#")
        cm_anchor = cm_abs_url[hash_pos:]
        cm_abs_url = cm_abs_url[:hash_pos]
    except Exception:
        return comment.get_absolute_url()

    if not page_number:
        page_number = 1
    else:
        try:
            page_number = int(page_number)
        except (TypeError, ValueError):
            raise PageNotAnInteger(_("That page number is not an integer"))

    if not comments_folded:
        comments_folded = ""
    else:
        try:
            comments_folded = {int(cid) for cid in comments_folded.split(",")}
        except (TypeError, ValueError):
            raise PageNotAnInteger(
                _("A comment ID in comments_folded is not an integer.")
            )

    qs_params = []

    if page_number > 1:
        cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
        qs_params.append(f"{cpage_qs_param}={page_number}")

    if len(comments_folded):
        cfold_qs_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM
        csv_of_cids = ",".join([str(cid) for cid in comments_folded])
        qs_params.append(f"{cfold_qs_param}={csv_of_cids}")

    if len(qs_params):
        return mark_safe(f"{cm_abs_url}?{'&'.join(qs_params)}{cm_anchor}")
    else:
        return f"{cm_abs_url}{cm_anchor}"


# ----------------------------------------------------------------------
class GetCommentsAPIPropsNode(Node):
    def __init__(self, obj):
        self.obj = Variable(obj)

    def render(self, context):
        obj = self.obj.resolve(context)
        user = context.get("user", None)
        request = context.get("request", None)
        props = frontend.comments_api_props(obj, user, request=request)
        return json.dumps(props)


@register.tag
def get_comments_api_props(parser, token):
    """
    Returns a JSON with properties required to use the REST API.

    See api.frontend.comments_props for full details on the props.

    Example::
        {% get_comments_api_props for comment %}
    """
    try:
        tag_name, args = token.contents.split(None, 1)
    except ValueError:
        raise TemplateSyntaxError(
            "%r tag requires arguments" % token.contents.split()[0]
        )
    match = re.search(r"for (\w+)", args)
    if not match:
        raise TemplateSyntaxError("%r tag had invalid arguments" % tag_name)
    obj = match.groups()[0]
    return GetCommentsAPIPropsNode(obj)


# ----------------------------------------------------------------------
@register.simple_tag
def comment_reaction_form_target(comment):
    """
    Get the target URL for the comment reaction form.

    Example::

        <form action="{% comment_reaction_form_target comment %}" method="post">
    """
    return reverse("comments-ink-react", args=(comment.id,))


class RenderCommentReactionsButtons(Node):
    def __init__(self, user_reactions):
        self.user_reactions = Variable(user_reactions)

    def render(self, context):
        context = {
            "reactions": get_comment_reactions_enum(),
            "user_reactions": self.user_reactions.resolve(context),
            "break_every": settings.COMMENTS_INK_REACTIONS_ROW_LENGTH,
        }
        template_list = [
            pth.format(theme_dir=theme_dir) for pth in _reactions_buttons_tmpl
        ]
        htmlstr = loader.render_to_string(template_list, context)
        return htmlstr


@register.tag
def render_comment_reactions_buttons(parser, token):
    """
    Renders template with reactions buttons, depending on the selected theme.

    Example usage::

        {% render_comment_reactions_buttons user_reactions %}

    Argument `user_reactions` is a list with `ReactionEnum` items, it
    contains a user's reactions to a comment. The template displays a button
    per each reaction returned from get_comment_reactions_enum().
    Each reaction that is already in the `user_reactions` list is marked as
    already clicked. This templatetag is used within the
    `react_to_comment.html` template.
    """
    try:
        _, args = token.contents.split(None, 1)
    except ValueError:
        raise TemplateSyntaxError(
            "%r tag requires argument" % token.contents.split()[0]
        )
    user_reactions = args
    return RenderCommentReactionsButtons(user_reactions)


@register.filter
def authors_list(cmt_reaction):
    """
    Returns a list with the result of applying the function
    COMMENTS_INK_API_USER_REPR to each authour of the given CommentReaction.
    """
    return [
        settings.COMMENTS_INK_API_USER_REPR(author)
        for author in cmt_reaction.authors.all()
    ]


@register.filter
def get_comment_reaction_enum(cmt_reaction):
    """
    Helper to get the ReactionEnum corresponding to given CommentReaction.
    """
    return get_comment_reactions_enum()(cmt_reaction.reaction)


# ----------------------------------------------------------------------
@register.simple_tag(takes_context=True)
def comment_css_thread_range(context, comment, prefix="l"):
    """
    Helper tag to return a string of CSS selectors that render vertical
    lines to represent comment threads. When comment's level matches the
    max_thread_level there is no vertical line as comments in the
    max_thread_level can not receivereplies.

    Returns a concatenated string of f'{prefix}{i}' for i in range(level + 1).
    When the given comment has level=2, and the maximum thread level is 2:

        `{% comment_css_thread_range comment %}`

    produces the string: "l0-mid l1-mid l2".
    """
    max_thread_level = context.get("max_thread_level", None)
    if not max_thread_level:
        ctype = ContentType.objects.get_for_model(comment.content_object)
        max_thread_level = max_thread_level_for_content_type(ctype)

    result = ""
    for i in range(comment.level + 1):
        if i == comment.level:
            if comment.level == max_thread_level:
                result += f"{prefix}{i} "
            else:
                result += f"{prefix}{i}-ini "
        else:
            result += f"{prefix}{i}-mid "
    return result.rstrip()


@register.filter(is_safe=True)
def reply_css_thread_range(level, prefix="l"):
    """
    Helper filter to return a string of CSS selectors that render vertical
    lines to represent comment threads. When comment level matches the
    max_thread_level there is no vertical line, as comments in the
    max_thread_level can not receive replies.

    Returns a concatenated string of f'{prefix}{i}' for i in range(level + 1).
    If the given comment object has level=1, using the filter as:

        `{{ comment.level|reply_css_thread_range }}`

    produces the string: "l0 l1".
    """
    result = ""
    for i in range(level + 1):
        result += f"{prefix}{i} "
    return mark_safe(result.rstrip())


@register.filter(is_safe=True)
def indent_divs(level, prefix="level-"):
    """
    Helper filter to return a concatenated string of divs for indentation.

    When called as {{ 2|indent_divs }} produces the string:

        '<div class="level-1"></div>
         <div class="level-2"></div>'
    """
    result = ""
    for i in range(1, level + 1):
        result += f'<div class="{prefix}{i}"></div>'
    return mark_safe(result)


@register.filter(is_safe=True)
def hline_div(level, prefix="line-"):
    """
    Helper filter to eeturns a DIV that renders a horizontal line connecting
    the vertical comment thread line with the comment reply box.

    When called as {{ comment.level|hline_div }} produces the string:

        '<div class="line-{comment.level}"></div>'
    """
    return mark_safe(f'<div class="{prefix}{level}"></div>')


@register.filter
def get_top_comment(reply_stack):
    """
    Helper filter used to list comments in the <comments/list.html> template.
    """
    return reply_stack[-1]


@register.filter
def pop_comments_gte(reply_stack, level_lte=0):
    """
    Helper filter used to list comments in the <comments/list.html> template.
    """
    comments_lte = []
    try:
        for index in range(len(reply_stack) - 1, -1, -1):
            if reply_stack[index].level < level_lte:
                break
            comments_lte.append(reply_stack.pop(index))
    finally:
        return comments_lte


@register.simple_tag(takes_context=True)
def push_comment(context, comment):
    """
    Helper filter used to list comments in the <comments/list.html> template.
    """
    context["reply_stack"].append(comment)
    return ""


@register.filter
def get_comment(comment_id: str):
    return get_model().objects.get(pk=int(comment_id))


@register.simple_tag()
def dci_custom_selector():
    return f"{settings.COMMENTS_INK_CSS_CUSTOM_SELECTOR}"


@register.simple_tag()
def get_dci_theme_dir():
    return theme_dir


@register.simple_tag(takes_context=True)
def get_folded_comments_without(context, comment):
    folded_comments = context.get("folded_comments", set())
    folded_comments.remove(comment.id)
    return ",".join([str(cm) for cm in folded_comments])


@register.simple_tag(takes_context=True)
def get_folded_comments_with(context, comment):
    folded_comments = context.get("folded_comments", set())
    folded_comments.add(comment.id)
    return ",".join([str(cm) for cm in folded_comments])


# ----------------------------------------------------------------------
@register.inclusion_tag("comments/only_users_can_post.html")
def render_only_users_can_post_template(object):
    return {"html_id_suffix": utils.get_html_id_suffix(object)}


class RenderCommentReactionsPanelTemplate(Node):
    def render(self, context):
        enums_details = [
            (enum.value, enum.label, enum.icon)
            for enum in get_comment_reactions_enum()
        ]
        context = {
            "enums_details": enums_details,
            "break_every": settings.COMMENTS_INK_REACTIONS_ROW_LENGTH,
        }
        template_list = [
            pth.format(theme_dir=theme_dir)
            for pth in _reactions_panel_template_tmpl
        ]
        htmlstr = loader.render_to_string(template_list, context)
        return htmlstr


@register.tag
def render_comment_reactions_panel_template(parser, token):
    return RenderCommentReactionsPanelTemplate()


# ----------------------------------------------------------------------
# Base class to handle object reactions.


class BaseObjectReactionsNode(Node):
    """
    Base helper class for handling tags:
     * render_object_reactions
     * render_object_reactions_form
    """

    @classmethod
    def handle_token(cls, parser, token):
        tokens = token.split_contents()
        if tokens[1] != "for":
            raise TemplateSyntaxError(
                "Second argument in %r tag must be 'for'" % tokens[0]
            )

        # {% render_object_reactions for obj %}
        if len(tokens) == 3:
            return cls(object_expr=parser.compile_filter(tokens[2]))

        # {% render_object_reactions for app.models pk %}
        elif len(tokens) == 4:
            return cls(
                ctype=BaseCommentNode.lookup_content_type(tokens[2], tokens[0]),
                object_pk_expr=parser.compile_filter(tokens[3]),
            )

    @staticmethod
    def lookup_content_type(token, tagname):
        try:
            app, model = token.split(".")
            return ContentType.objects.get_by_natural_key(app, model)
        except ValueError:
            raise TemplateSyntaxError(
                "Third argument in %r must be in the format 'app.model'"
                % tagname
            )
        except ContentType.DoesNotExist:
            raise TemplateSyntaxError(
                "%r tag has non-existant content-type: '%s.%s'"
                % (tagname, app, model)
            )

    def __init__(self, ctype=None, object_pk_expr=None, object_expr=None):
        if ctype is None and object_expr is None:
            raise TemplateSyntaxError(
                "Template tag must be given either a literal object "
                "or a ctype and object pk."
            )
        self.ctype = ctype
        self.object_pk_expr = object_pk_expr
        self.object_expr = object_expr

    def resolve_ctype_and_object_pk(self, context):
        """
        Assign values to self attributes ctype, object_pk and site_id.
        """
        self.object_pk = None
        self.site_id = None

        request = context.get("request", None)
        self.site_id = utils.get_current_site_id(request)

        if self.object_expr:
            try:
                obj = self.object_expr.resolve(context)
            except VariableDoesNotExist:
                return
            else:
                self.ctype = ContentType.objects.get_for_model(obj)
                self.object_pk = obj.pk
        else:
            self.object_pk = self.object_pk_expr.resolve(
                context, gnore_failures=True
            )

        app_model = "%s.%s" % (self.ctype.app_label, self.ctype.model)
        options = utils.get_app_model_options(content_type=app_model)
        if not "object_reactions_enabled" in options:
            raise Exception(
                "Object reactions for '%s' is not enabled in the setting "
                "COMMENTS_INK_APP_MODEL_OPTIONS in the settings module"
                % app_model
            )

    @property
    def template_list(self):
        return [
            pth.format(
                theme_dir=theme_dir,
                app_label=self.ctype.app_label,
                model=self.ctype.model,
            )
            for pth in self.tag_templates
        ]

    def get_object_reactions(self):
        """Returns list of dicts with object reactions and their counters."""
        reactionsd = dict(
            [
                (item.reaction, item.counter)
                for item in ObjectReaction.objects.filter(
                    content_type=self.ctype,
                    object_pk=self.object_pk,
                    site__id=self.site_id,
                )
            ]
        )

        object_reactions = []
        for item in get_object_reactions_enum():
            if item.value in reactionsd:
                counter = reactionsd[item.value]
            else:
                counter = 0
            object_reactions.append(
                {
                    "value": item.value,
                    "label": item.label,
                    "icon": item.icon,
                    "counter": counter,
                }
            )

        return object_reactions

    def render(self, context):
        request = context.get("request", None)
        try:
            self.resolve_ctype_and_object_pk(context)
            object = self.ctype.get_object_for_this_type(pk=self.object_pk)
        except AttributeError:
            return ""

        cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
        page = (request and request.GET.get(cpage_qs_param, None)) or 1

        cfold_qs_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM
        cfold = (request and request.GET.get(cfold_qs_param, None)) or None
        fold = (cfold and {int(cid) for cid in cfold.split(",")}) or {}

        context = {
            "object": object,
            "object_reactions": self.get_object_reactions(),
            "comments_page_qs_param": cpage_qs_param,
            cpage_qs_param: page,
            "comments_fold_qs_param": cfold_qs_param,
            cfold_qs_param: ",".join([str(cid) for cid in fold]),
        }
        htmlstr = loader.render_to_string(self.template_list, context, request)
        return htmlstr


class RenderObjectReactions(BaseObjectReactionsNode):
    """Template tag to render reactions posted to a given object."""

    tag_templates = _object_reactions_tmpl


class RenderObjectReactionsForm(BaseObjectReactionsNode):
    """Template tag to render the object reactions form."""

    tag_templates = _object_reactions_form_tmpl


@register.tag
def render_object_reactions(parser, token):
    """
    Render the object reactions list through the ``object_reactions.html``
    template. The template is looked for in the following directories:
     * comments/<app_label>/<model>/
     * comments/<app_label>/
     * comments/
    If the setting `COMMENTS_INK_THEME_DIR` is provided in the settings module
    the previous list contains is preceded by the following list:
     * comments/<theme_dir>/<app_label>/<model>/
     * comments/<theme_dir>/<app_label>/
     * comments/<theme_dir>/

    Syntax::

        {% render_object_reactions for [object] %}
        {% render_object_reactions for [app].[model] [obj_id] %}

    Example usage::

        {% render_object_reactions for post %}

    """
    return RenderObjectReactions.handle_token(parser, token)


@register.tag
def render_object_reactions_form(parser, token):
    """
    Render object reactions using the ``object_reactions_form.html`` template.
    The template is looked for in the following directories:
     * comments/<app_label>/<model>/
     * comments/<app_label>/
     * comments/
    If the setting `COMMENTS_INK_THEME_DIR` is provided in the settings module
    the previous list contains is preceded by the following list:
     * comments/<theme_dir>/<app_label>/<model>/
     * comments/<theme_dir>/<app_label>/
     * comments/<theme_dir>/

    Syntax::

        {% render_object_reactions_form for [object] %}
        {% render_object_reactions_form for [app].[model] [obj_id] %}

    Example usage::

        {% render_object_reactions_form for post %}

    """
    return RenderObjectReactionsForm.handle_token(parser, token)


@register.simple_tag
def object_reactions_form_target(object):
    """
    Get the target URL for the object reaction form.

    Example::

        <form action="{% object_reactions_form_target object %}" method="post">
    """
    ctype = ContentType.objects.get_for_model(object)
    return reverse("comments-ink-object-react", args=(ctype.id, object.id))


class RenderObjectReactionsButtons(Node):
    def __init__(self, user_reactions):
        self.user_reactions = Variable(user_reactions)

    def render(self, context):
        context = {
            "reactions": get_object_reactions_enum(),
            "user_reactions": self.user_reactions.resolve(context),
        }
        template_list = [
            pth.format(theme_dir=theme_dir)
            for pth in _object_reactions_buttons_tmpl
        ]
        htmlstr = loader.render_to_string(template_list, context)
        return htmlstr


@register.tag
def render_object_reactions_buttons(parser, token):
    """
    Renders template with reactions buttons, depending on the selected theme.

    Example usage::

        {% render_object_reactions_buttons user_reactions %}

    Argument `user_reactions` is a list with `ReactionEnum` items, it
    contains a user's reactions to an object. The template displays a button
    per each reaction returned from get_object_reactions_enum().
    Each reaction that is already in the `user_reactions` list is marked as
    already clicked. This templatetag is used within the `react_to_object.html`
    template.
    """
    try:
        _, args = token.contents.split(None, 1)
    except ValueError:
        raise TemplateSyntaxError(
            "%r tag requires argument" % token.contents.split()[0]
        )
    user_reactions = args
    return RenderObjectReactionsButtons(user_reactions)


# ----------------------------------------------------------------------
# Template tag for themes 'avatar_in_thread' and 'avatar_in_header'


def get_gravatar_url(email, size=48, avatar="identicon"):
    """
    This is the parameter of the production avatar.
    The first parameter is the way of generating the
    avatar and the second one is the size.
    The way os generating has mp/identicon/monsterid/wavatar/retro/hide.
    """
    return "//www.gravatar.com/avatar/%s?%s&d=%s" % (
        hashlib.md5(email.lower().encode("utf-8")).hexdigest(),
        urlencode({"s": str(size)}),
        avatar,
    )


# ----------------------------------------------------------------------
# The following code requires django-avatar.

if apps.is_installed("avatar"):  # pragma: no cover

    from avatar.templatetags.avatar_tags import avatar
    from avatar.utils import cache_result

    @cache_result
    @register.simple_tag
    def get_user_avatar_or_gravatar(email, config="48,identicon"):
        size, gravatar_type = config.split(",")
        try:
            size_number = int(size)
        except ValueError:
            raise Http404(_("The given size is not a number: %s" % repr(size)))

        try:
            user = get_user_model().objects.get(email=email)
            return avatar(user, size_number)
        except get_user_model().DoesNotExist:
            url = get_gravatar_url(email, size_number, gravatar_type)
            return mark_safe(
                '<img src="%s" height="%d" width="%d">'
                % (url, size_number, size_number)
            )

else:

    @register.simple_tag
    def get_user_avatar_or_gravatar(*args, **kwargs):
        raise TemplateSyntaxError(
            "The 'get_user_avatar_or_gravatar' template tag requires "
            "to have installed the package 'django-avatar'."
        )



"""
Pagination for comments threads.

Given the following comments structure the paginator will display comments
in several pages containing different number of comments, so that comment
threads (a thread of comments have all the same value for thread_id) are
fully displayed in one page. Even if the number of comments in the thread
exceed the `per_page` value.

Relevant settings:
 * COMMENTS_INK_ITEMS_PER_PAGE = 25
 * COMMENTS_INK_MAX_LAST_PAGE_ORPHANS = 10

Example 1:
  Comment IDs of level 0:  [ 1,  2,  3,  4,  5,  6,  7,  8]
  Their `nested_count`:    [10, 10, 10, 10, 10, 10,  5,  4]
  Computed comments/page:  [    22,     22,             33]

  Thread 1 has 11 comments, still below per_page. We add Comment thread 2, now
  we have 22 comments, still below per_page. We add Comment thread 3 and now
  the result exceeds per_page, so we will only accept it if not only adding
  Comment thread 3 but all the rest represents a number of comments below
  per_page + orphans, which is not the case: 77 comments > per_page + orphans.
  So page 1 contains 22 comments, and we move to compute comments for page 2.
  Page 2 start with 11 comments of Thread 3, we add 11 comments of Thread 4,
  and when we try to add Thread 5 we have over per_page, so we will only accept
  it if adding all the rest will represent a number of comments below
  per_page + orphans, which is not the case: 55 > 35, so page 2 will contain
  22 comments of Thread 3 and Thread 4. Page 3 starts with 11 comments of
  Thread 5 + 11 comments of Thread 6. Adding the 6 comments of Thread 7 would
  exceed per_page, so we will only accept to change page 3 if adding all the
  rest would keep the number of comments below per_page + orphans. And that is
  the case now: 22 comments (thread 5 + thread 6) + 11 (thread 7 + thread 8) is
  33 comments, which is below 35, so the last page contains 33 comments.

Example 2:
  Comment IDs of level 0:  [ 1,  2, 3]
  Their `nested_count`:    [24, 24, 8]
  Computed comments/page:  [25,    34]

  Comment 1 has 24 nested comments, it doesn't exceed per_page (25), but
  adding all the rest of comment threads would exceed per_page + orphans, so
  the 1st page displays 25 comments (comment 1 + its 24 nested comments).
  Comment thread 2 is made of 25 comments. Adding all the rest (34) would not
  exceed per_page + orphans (35), so the 2nd (or last page) will have 34
  comments.

Example 3:
  Comment IDs of level 0:  [ 1,  2, 3]
  Their `nested_count`:    [24,  8, 8]
  Computed comments/page:  [25,    18]

  Comment 1 has 24 nested comments, it doesn't exceed per_page (25). Adding
  all the rest (18) would represent 43 comments which exceed per_page + orphans
  (35). So the 1st page displays the 25 comments of thread 1. For page 2,
  comment thread 2 has 9 comments, and adding the rest is still below
  per_page + orphans, so the 2nd page will contain 18 comments.

"""
import logging

from django.core.paginator import Paginator
from django.db.models.query import QuerySet
from django.utils.functional import cached_property
from django_comments_ink import caching

logger = logging.getLogger(__name__)


class CommentsPaginator(Paginator):
    """
    Paginator for comments.

    Given an `object_list` of comments, it checks how many threads fit per page.
    If a thread has more comments than the given `per_page` limit it ignores the
    limit so that the thread is not cut in several pages.
    """

    def __init__(self, *args, **kwargs):
        self.comments_folded = kwargs.pop("comments_folded", {})
        self.ckey_prefix = kwargs.pop("cache_key_prefix", "")
        self.dci_cache = caching.get_cache()

        # Comment out next lines when debugging paginator's cache keys.
        if self.dci_cache != None and self.ckey_prefix != "":
            sub_keys_set = self.dci_cache.get(self.ckey_prefix)
            if sub_keys_set:
                logger.debug(
                    "sub_keys_set for %s: %s", self.ckey_prefix, sub_keys_set
                )

        super().__init__(*args, **kwargs)
        if type(self.object_list) is not QuerySet:
            raise TypeError("'object_list' is not a QuerySet.")

    def get_subkey_cache(self, subkey):
        if self.dci_cache == None or self.ckey_prefix == "":
            return
        # If the key <sub_ckey> does exist as a key in
        # the set stored in the <ckey_prefix> in the cache, then
        # access the combinaned <self.ckey_prefix>/<sub_ckey>
        # to get the previously computed object_list.
        sub_keys_set = self.dci_cache.get(self.ckey_prefix)
        if sub_keys_set and subkey in sub_keys_set:
            composed_key = f"{self.ckey_prefix}/{subkey}"
            result = self.dci_cache.get(composed_key)
            if result != None:
                return result

    def set_subkey_cache(self, subkey, value):
        if self.dci_cache == None or self.ckey_prefix == "":
            return
        # Save the object_list in cache.
        sub_keys_set = self.dci_cache.get(self.ckey_prefix)
        if sub_keys_set != None:
            if not subkey in sub_keys_set:
                sub_keys_set.add(subkey)
        else:
            sub_keys_set = {subkey}
        logger.debug("Caching key %s, value %s", self.ckey_prefix, sub_keys_set)
        self.dci_cache.set(self.ckey_prefix, sub_keys_set, timeout=None)

        # Store the object_list in cache using a composed key.
        composed_key = f"{self.ckey_prefix}/{subkey}"
        logger.debug("Adding %s to the cache", composed_key)
        self.dci_cache.set(composed_key, value, timeout=None)

    def get_count_in_thread(self, comment):
        if self.comments_folded and comment.id in self.comments_folded:
            return 1
        else:
            return comment.nested_count + 1

    def get_sub_ckey(self, page, fold):
        page_part = "all-pages" if page == None else f"page-{str(page)}"
        if fold:
            fold_part = f"folded-{','.join([str(cid) for cid in fold])}"
        else:
            fold_part = "all-unfolded"
        return page_part + "-" + fold_part

    @cached_property
    def in_page(self):
        """
        Calculate the variable number of comments displayed in each page.

        Returns a list. Each index item represents the number of comments to
        display in the page index + 1.
        """
        inpage_subkey = (
            self.get_sub_ckey(None, self.comments_folded) + ":in_page_list"
        )
        inpage = self.get_subkey_cache(inpage_subkey)
        if inpage != None:
            logger.debug("in_page from cache %s: %s", self.ckey_prefix, inpage)
            return inpage

        ptotal = 0  # Page total number of comments.
        inpage = []

        cgroups = [
            self.get_count_in_thread(cm)
            for cm in self.object_list.filter(level=0)
        ]
        for index, group_count in enumerate(cgroups):
            if ptotal > 0 and ptotal + group_count > self.per_page:
                if sum(cgroups[index:], ptotal) > self.per_page + self.orphans:
                    # All comments are too many to be in this page.
                    inpage.append(ptotal)
                    ptotal = group_count
                else:
                    inpage.append(sum(cgroups[index:], ptotal))
                    ptotal = 0
                    break
            else:
                ptotal += group_count
        if ptotal:
            inpage.append(ptotal)
        self.set_subkey_cache(inpage_subkey, inpage)  # Store it in cache.
        logger.debug(
            "in_page computed (saved in cache) %s: %s", self.ckey_prefix, inpage
        )
        return inpage

    def page(self, number):
        number = self.validate_number(number)

        # See whether the object_list has been stored in the cache.
        sub_ckey = self.get_sub_ckey(number, self.comments_folded)
        object_list = self.get_subkey_cache(sub_ckey)
        if object_list != None:
            return self._get_page(object_list, number, self)

        if number == 1:
            bottom = 0
        else:
            bottom = sum(self.in_page[0 : number - 1])
        top = bottom + self.in_page[number - 1]
        object_list = self.object_list[bottom:top]

        # Store it in cache.
        self.set_subkey_cache(sub_ckey, object_list)

        return self._get_page(object_list, number, self)

    @cached_property
    def count(self):
        subkey = "count"
        result = self.get_subkey_cache(subkey)
        if result != None:
            return result
        count = super().count
        self.set_subkey_cache(subkey, count)  # Store it in cache.
        return count

    @cached_property
    def num_pages(self):
        """Return the total number of pages."""
        if self.count == 0 and not self.allow_empty_first_page:
            return 0
        return len(self.in_page)



from django.core.management.base import BaseCommand
from django.db import connections
from django.db.utils import ConnectionDoesNotExist, IntegrityError
from django_comments.models import Comment
from django_comments_ink.models import InkComment

__all__ = ["Command"]


class Command(BaseCommand):
    help = "Load the inkcomment table with valid data from django_comments."

    def add_arguments(self, parser):
        parser.add_argument("using", nargs="*", type=str)

    def populate_db(self, cursor):
        for comment in Comment.objects.all():
            sql = (
                "INSERT INTO %(table)s "
                "       ('comment_ptr_id', 'thread_id', 'parent_id',"
                "        'level', 'order', 'followup', 'nested_count') "
                "VALUES (%(id)d, %(id)d, %(id)d, 0, 1, FALSE, 0)"
            )
            cursor.execute(
                sql % {"table": InkComment._meta.db_table, "id": comment.id}
            )

    def handle(self, *args, **options):
        total = 0
        using = options["using"] or ["default"]
        for db_conn in using:
            try:
                self.populate_db(connections[db_conn].cursor())
                total += InkComment.objects.using(db_conn).count()
            except ConnectionDoesNotExist:
                self.stdout.write(
                    "DB connection '%s' does not exist." % db_conn
                )
            except IntegrityError:
                if db_conn != "default":
                    self.stdout.write(
                        "Table '%s' (in '%s' DB connection) "
                        "must be empty." % (InkComment._meta.db_table, db_conn)
                    )
                else:
                    self.stdout.write(
                        "Table '%s' must be empty." % InkComment._meta.db_table
                    )
            finally:
                continue
        self.stdout.write("Added %d InkComment object(s)." % total)



from pathlib import Path

from setuptools import find_packages, setup

BASE_DIR = Path(__file__).resolve().parent


readme = ""
with open(BASE_DIR / "README.md", "r") as readme_file:
    readme = readme_file.read()


requirements = []
with open(BASE_DIR / "requirements.in", "r") as req_file:
    for item in req_file:
        requirements.append(item)

setup(
    name="django-comments-ink",
    version="0.0.1",
    packages=find_packages(),
    include_package_data=True,
    license="MIT",
    description=(
        "Django pluggable commenting app with comment threads, follow-up "
        "notifications, mail confirmation, comment reactions and votes, and "
        "comment moderation. It supersedes django-comments-xtd."
    ),
    long_description=readme,
    long_description_content_type="text/x-rst",
    author="Daniela Rus Morales",
    author_email="danirus@eml.cc",
    maintainer="Daniela Rus Morales",
    maintainer_email="danirus@eml.cc",
    url="https://github.com/comments-ink/django-comments-ink",
    install_requires=requirements,
    setup_requires=["wheel"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Framework :: Django",
        "Natural Language :: English",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: News/Diary",
    ],
    test_suite="dummy",
    zip_safe=True,
)



from django.apps import apps
from django.contrib.sites.shortcuts import get_current_site
from django.utils import formats, timezone
from django.utils.html import escape
from django.utils.translation import activate, get_language
from django.utils.translation import gettext as _
from django_comments import get_form
from django_comments.forms import CommentSecurityForm
from django_comments.models import CommentFlag
from django_comments.signals import comment_was_posted, comment_will_be_posted
from django_comments_ink import (
    get_comment_reactions_enum,
    get_model,
    signed,
    views,
)
from django_comments_ink.conf import settings
from django_comments_ink.models import (
    CommentReaction,
    InkComment,
    TmpInkComment,
    max_thread_level_for_content_type,
)
from django_comments_ink.signals import (
    confirmation_received,
    should_request_be_authorized,
)
from django_comments_ink.utils import get_app_model_options
from rest_framework import exceptions, serializers

COMMENT_MAX_LENGTH = getattr(settings, "COMMENT_MAX_LENGTH", 3000)


class WriteCommentSerializer(serializers.Serializer):
    content_type = serializers.CharField()
    object_pk = serializers.CharField()
    timestamp = serializers.CharField(required=False)
    security_hash = serializers.CharField(required=False)
    honeypot = serializers.CharField(required=False, allow_blank=True)
    name = serializers.CharField(allow_blank=True)
    email = serializers.EmailField(allow_blank=True)
    url = serializers.URLField(required=False)
    comment = serializers.CharField(max_length=COMMENT_MAX_LENGTH)
    followup = serializers.BooleanField(default=False)
    reply_to = serializers.IntegerField(default=0)

    def __init__(self, *args, **kwargs):
        self.request = kwargs["context"]["request"]
        super(WriteCommentSerializer, self).__init__(*args, **kwargs)

    def validate_name(self, value):
        if value.strip():
            return value.strip()
        if self.request.user.is_authenticated:
            name = None
            if hasattr(self.request.user, "get_full_name"):
                name = self.request.user.get_full_name()
            elif hasattr(self.request.user, "get_username"):
                name = self.request.user.get_username()
            if name:
                return name
        raise serializers.ValidationError("This field is required")

    def validate_email(self, value):
        if value.strip():
            return value.strip()
        if self.request.user.is_authenticated:
            UserModel = apps.get_model(settings.AUTH_USER_MODEL)
            if hasattr(UserModel, "get_email_field_name"):
                email_field = UserModel.get_email_field_name()
                email = getattr(self.request.user, email_field, None)
                if email:
                    return email
        raise serializers.ValidationError("This field is required")

    def validate_reply_to(self, value):
        if value != 0:
            try:
                parent = get_model().objects.get(pk=value)
            except get_model().DoesNotExist:
                raise serializers.ValidationError(
                    "reply_to comment does not exist"
                )
            else:
                max = max_thread_level_for_content_type(parent.content_type)
                if parent.level == max:
                    raise serializers.ValidationError(
                        "Max thread level reached"
                    )
        return value

    def validate(self, data):
        ctype = data.get("content_type")
        object_pk = data.get("object_pk")
        try:
            model = apps.get_model(*ctype.split(".", 1))
            target = model._default_manager.get(pk=object_pk)
            whocan = get_app_model_options(content_type=ctype)["who_can_post"]
        except (AttributeError, TypeError, LookupError):
            raise serializers.ValidationError(
                "Invalid content_type value: %r" % escape(ctype)
            )
        except model.DoesNotExist:
            raise serializers.ValidationError(
                "No object matching content-type %r and object PK %r."
                % (escape(ctype), escape(object_pk))
            )
        except (ValueError, serializers.ValidationError) as e:
            raise serializers.ValidationError(
                "Attempting to get content-type %r and object PK %r "
                "raised %s"
                % (escape(ctype), escape(object_pk), e.__class__.__name__)
            )
        else:
            if whocan == "users" and not self.request.user.is_authenticated:
                raise serializers.ValidationError("User not authenticated")

        # Signal that the request allows to be authorized.
        responses = should_request_be_authorized.send(
            sender=target.__class__, comment=target, request=self.request
        )

        for (receiver, response) in responses:
            if response is True:
                # A positive response indicates that the POST request
                # must be trusted. So inject the CommentSecurityForm values
                # to pass the form validation step.
                secform = CommentSecurityForm(target)
                data.update(
                    {
                        "honeypot": "",
                        "timestamp": secform["timestamp"].value(),
                        "security_hash": secform["security_hash"].value(),
                    }
                )
                break

        self.form = get_form()(target, data=data)
        if not self.form.is_valid():
            # Check only security information as the rest
            # of the fields have already been validated.
            if self.form.security_errors():
                raise exceptions.PermissionDenied()
        return data

    def save(self):
        # resp object is a dictionary. The code key indicates the possible
        # four states the comment can be in:
        #  * Comment created (http 201),
        #  * Confirmation sent by mail (http 204),
        #  * Comment in moderation (http 202),
        #  * Comment rejected (http 403),
        #  * Comment have bad data (http 400).
        site = get_current_site(self.request)
        resp = {
            "code": -1,
            "comment": self.form.get_comment_object(site_id=site.id),
        }
        resp["comment"].ip_address = self.request.META.get("REMOTE_ADDR", None)

        if self.request.user.is_authenticated:
            resp["comment"].user = self.request.user

        # Signal that the comment is about to be saved
        responses = comment_will_be_posted.send(
            sender=TmpInkComment, comment=resp["comment"], request=self.request
        )
        for (receiver, response) in responses:
            if response is False:
                resp["code"] = 403  # Rejected.
                return resp

        # Replicate logic from django_comments_ink.views.on_comment_was_posted.
        if (
            not settings.COMMENTS_INK_CONFIRM_EMAIL
            or self.request.user.is_authenticated
        ):
            if views._get_comment_if_exists(resp["comment"]) is None:
                new_comment = views._create_comment(resp["comment"])
                resp["comment"].ink_comment = new_comment
                confirmation_received.send(
                    sender=TmpInkComment,
                    comment=resp["comment"],
                    request=self.request,
                )
                comment_was_posted.send(
                    sender=new_comment.__class__,
                    comment=new_comment,
                    request=self.request,
                )
                if resp["comment"].is_public:
                    resp["code"] = 201
                    views.notify_comment_followers(new_comment)
                else:
                    resp["code"] = 202
        else:
            key = signed.dumps(
                resp["comment"],
                compress=True,
                extra_key=settings.COMMENTS_INK_SALT,
            )
            views.send_email_confirmation_request(resp["comment"], key, site)
            resp["code"] = 204  # Confirmation sent by mail.

        return resp


class FlagSerializer(serializers.ModelSerializer):
    flag_choices = {"report": CommentFlag.SUGGEST_REMOVAL}

    class Meta:
        model = CommentFlag
        fields = (
            "comment",
            "flag",
        )

    def validate(self, data):
        # Validate flag.
        if data["flag"] not in self.flag_choices:
            raise serializers.ValidationError("Invalid flag.")
        data["flag"] = self.flag_choices[data["flag"]]
        return data


class ReadReactionsField(serializers.RelatedField):
    def to_representation(self, value):
        reaction_item = get_comment_reactions_enum()(value.reaction)
        return {
            "reaction": value.reaction,
            "label": reaction_item.label,
            "icon": reaction_item.icon,
            "counter": value.counter,
            "authors": [
                {
                    "id": author.id,
                    "author": settings.COMMENTS_INK_API_USER_REPR(author),
                }
                for author in value.authors.all()
            ],
        }


class ReadCommentSerializer(serializers.ModelSerializer):
    user_name = serializers.CharField(max_length=50, read_only=True)
    user_url = serializers.CharField(read_only=True)
    submit_date = serializers.SerializerMethodField()
    parent_id = serializers.IntegerField(default=0, read_only=True)
    level = serializers.IntegerField(read_only=True)
    is_removed = serializers.BooleanField(read_only=True)
    comment = serializers.SerializerMethodField()
    allow_reply = serializers.SerializerMethodField()
    permalink = serializers.SerializerMethodField()
    flags = serializers.SerializerMethodField()
    reactions = ReadReactionsField(many=True, read_only=True)

    class Meta:
        model = InkComment
        fields = (
            "id",
            "user_name",
            "user_url",
            "permalink",
            "comment",
            "submit_date",
            "parent_id",
            "level",
            "is_removed",
            "allow_reply",
            "flags",
            "reactions",
        )

    def __init__(self, *args, **kwargs):
        self.request = kwargs["context"]["request"]
        super(ReadCommentSerializer, self).__init__(*args, **kwargs)

    def get_submit_date(self, obj):
        activate(get_language())
        if settings.USE_TZ:
            submit_date = timezone.localtime(obj.submit_date)
        else:
            submit_date = obj.submit_date
        return formats.date_format(
            submit_date, "DATETIME_FORMAT", use_l10n=True
        )

    def get_comment(self, obj):
        if obj.is_removed:
            return _("This comment has been removed.")
        else:
            return obj.comment

    def get_allow_reply(self, obj):
        return obj.allow_thread()

    def get_permalink(self, obj):
        return obj.get_absolute_url()

    def get_flags(self, obj):
        qs = obj.flags.filter(flag=CommentFlag.SUGGEST_REMOVAL)
        flags = []
        for flag in qs:
            flags.append(
                {
                    "flag": "removal",
                    "user": settings.COMMENTS_INK_API_USER_REPR(flag.user),
                    "id": flag.user.id,
                }
            )
        return flags


class WriteCommentReactionSerializer(serializers.ModelSerializer):
    class Meta:
        model = CommentReaction
        fields = (
            "reaction",
            "comment",
        )



from django.apps import AppConfig
from django.db.models.signals import pre_save


class CommentsInkConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_comments_ink"
    verbose_name = "Comments Ink"

    def ready(self):
        from django_comments_ink import get_model
        from django_comments_ink.conf import settings
        from django_comments_ink.models import publish_or_withhold_on_pre_save

        if getattr(settings, "COMMENTS_HIDE_REMOVED", True) or getattr(
            settings, "COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED", True
        ):
            model_app_label = get_model()._meta.label
            pre_save.connect(
                publish_or_withhold_on_pre_save, sender=model_app_label
            )
